from __future__ import annotations

import json
import math
import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Iterable

from PySide6.QtCore import Qt, QTimer, Signal as pyqtSignal
from PySide6.QtGui import QKeyEvent, QMouseEvent
from PySide6.QtWidgets import (
    QAbstractItemView,
    QCheckBox,
    QComboBox,
    QDialog,
    QDoubleSpinBox,
    QFileDialog,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from astropy import units as u
from astropy.wcs.utils import proj_plane_pixel_scales

from takefits.core.marker import (
    Marker,
    MarkerState,
    MarkerStyle,
    LineMarker,
    SymbolMarker,
    TextMarker,
    marker_states_to_ds9,
    marker_states_to_json,
)
from takefits.core.marker_manager import MarkerManager
from takefits.core.wcs_frames import (
    available_display_frames,
    display_axis_type,
    display_frame_label,
    frame_is_available,
    normalize_display_frame,
    plane_axis_indices,
    plane_inputs_to_native,
    plane_values_for_display,
)



class DeselectableListWidget(QListWidget):
    delete_pressed = pyqtSignal()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)
        if self.itemAt(event.pos()) is None:
            self.clearSelection()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        super().keyPressEvent(event)
        if event.key() in (Qt.Key.Key_Delete, Qt.Key.Key_Backspace):
            self.delete_pressed.emit()



@dataclass
class SymbolControls:
    widget: QWidget
    size_spin: QDoubleSpinBox
    linewidth_spin: QDoubleSpinBox
    edge_color_label: QLabel
    edge_color_combo: QComboBox


@dataclass
class LineControls:
    widget: QWidget
    length_spin: QDoubleSpinBox
    unit_combo: QComboBox
    angle_spin: QDoubleSpinBox
    linewidth_spin: QDoubleSpinBox
    style_combo: QComboBox


@dataclass
class TextControls:
    widget: QWidget
    font_size_spin: QDoubleSpinBox
    rotation_spin: QDoubleSpinBox
    font_combo: QComboBox


class MarkerPanel(QDialog):
    """Dialog for managing draw markers, lines, and annotations across planes."""

    COLOR_OPTIONS = [
        "None",
        "blue",
        "red",
        "green",
        "cyan",
        "magenta",
        "black",
        "white",
        "gray",
        "orange",
        "purple",
        "yellow",
        "yellowgreen",
        "olive",
        "lime",
    ]

    UNIT_OPTIONS = ["pixel", "arcsec", "arcmin", "deg"]

    FONT_OPTIONS = [
        "Arial",
        "DejaVu Sans",
        "STIXGeneral",
        "Helvetica",
        "Times New Roman",
        "Courier New",
        "Monaco",
        "Verdana",
    ]

    TYPE_OPTIONS = [
        ("Circle", {"kind": "symbol", "symbol": "o", "glyph": "●"}),
        ("Square", {"kind": "symbol", "symbol": "s", "glyph": "■"}),
        ("Diamond", {"kind": "symbol", "symbol": "D", "glyph": "◆"}),
        ("Triangle Up", {"kind": "symbol", "symbol": "^", "glyph": "▲"}),
        ("Triangle Down", {"kind": "symbol", "symbol": "v", "glyph": "▼"}),
        ("Star", {"kind": "symbol", "symbol": "*", "glyph": "★"}),
        ("Plus", {"kind": "symbol", "symbol": "+", "glyph": "+"}),
        ("Cross", {"kind": "symbol", "symbol": "x", "glyph": "✕"}),
        ("Line", {"kind": "line", "glyph": "―"}),
        ("Text", {"kind": "text"}),
    ]

    def __init__(self, viewer, marker_manager: MarkerManager):
        super().__init__(viewer)
        self.viewer = viewer
        self.marker_manager = marker_manager
        self._current_plane = getattr(viewer, "plane", "xy")
        default_plane_getter = getattr(viewer, "default_marker_plane", None)
        if callable(default_plane_getter):
            try:
                default_plane = default_plane_getter()
            except Exception:
                default_plane = None
            if default_plane:
                self._current_plane = str(default_plane)
        self._current_marker_id: Optional[str] = None
        self._suspend_updates = False
        self._suspend_selection_sync = False
        self._pending_selection_id: Optional[str] = None
        self._selection_from_list = False
        self.property_rows = {"symbol": [], "line": [], "text": []}
        self.link_all_checkbox: Optional[QCheckBox] = None
        self._last_type_kind: Optional[str] = None
        self._last_text_value: str = "Text"
        self._suppress_label_from_text: bool = False
        self.detail_group: Optional[QGroupBox] = None
        self._detail_group_heights: Dict[str, int] = {}
        self._detail_group_fixed_height: Optional[int] = None

        self.setWindowTitle(self._window_title_for_viewer())
        self.setModal(False)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)

        self._build_ui()
        self._connect_signals()
        self._precompute_detail_group_heights()
        self._on_type_changed(self.type_combo.currentIndex())
        self.placement_toggle.setChecked(True)
        self._configure_placement()
        self._refresh_marker_list()

    def _window_title_for_viewer(self) -> str:
        viewer = getattr(self, "viewer", None)
        class_name = str(getattr(viewer, "__class__", type(viewer)).__name__ or "").lower()
        if class_name == "channelmapwindow":
            return "Markers (Channel map)"
        if class_name == "integresultwindow":
            return "Markers (Integ)"
        if class_name in {"mainwindow", "subwindow"}:
            return "Markers (Main)"
        return "Markers"

    # ------------------------------------------------------------------
    # UI construction
    def _build_ui(self) -> None:
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(10, 10, 10, 10)
        main_layout.setSpacing(8)

        type_row = QHBoxLayout()
        type_row.setSpacing(6)
        type_row.addWidget(QLabel("Type", self))

        self.type_combo = QComboBox(self)
        for label, data in self.TYPE_OPTIONS:
            glyph = data.get("glyph")
            if glyph:
                display = f"{label} {glyph}"
            else:
                display = label
            self.type_combo.addItem(display)
        type_row.addWidget(self.type_combo, stretch=1)

        self.placement_toggle = QCheckBox("Placement mode", self)
        type_row.addWidget(self.placement_toggle)
        if self._viewer_supports_link_all():
            self.link_all_checkbox = QCheckBox("All tiles", self)
            self.link_all_checkbox.setToolTip("Place markers on every channel tile")
            self.link_all_checkbox.setChecked(self._viewer_link_all_enabled())
            type_row.addWidget(self.link_all_checkbox)
        type_row.addStretch()
        main_layout.addLayout(type_row)

        detail_group = QGroupBox("Properties", self)
        self.detail_group = detail_group
        self.properties_form = QFormLayout(detail_group)
        self.properties_form.setContentsMargins(8, 8, 8, 8)
        self.properties_form.setVerticalSpacing(4)

        self.pixel_x_spin = self._make_coordinate_spin(detail_group)
        self.pixel_y_spin = self._make_coordinate_spin(detail_group)
        self.world_x_edit = self._make_world_line_edit(detail_group, "World X")
        self.world_y_edit = self._make_world_line_edit(detail_group, "World Y")
        self.world_frame_combo = QComboBox(detail_group)
        self._populate_world_frame_combo()
        self.pixel_x_spin.setFixedWidth(110)
        self.pixel_y_spin.setFixedWidth(110)
        if self.pixel_x_spin.lineEdit() is not None:
            self.pixel_x_spin.lineEdit().returnPressed.connect(self._on_pixel_entry_return_pressed)
        if self.pixel_y_spin.lineEdit() is not None:
            self.pixel_y_spin.lineEdit().returnPressed.connect(self._on_pixel_entry_return_pressed)
        self.world_x_edit.returnPressed.connect(self._on_world_entry_return_pressed)
        self.world_y_edit.returnPressed.connect(self._on_world_entry_return_pressed)

        x_row = QHBoxLayout()
        x_row.setContentsMargins(0, 0, 0, 0)
        x_row.setSpacing(4)
        x_row.addWidget(self.pixel_x_spin)
        x_row.addWidget(self.world_x_edit)
        x_widget = QWidget(detail_group)
        x_widget.setLayout(x_row)

        y_row = QHBoxLayout()
        y_row.setContentsMargins(0, 0, 0, 0)
        y_row.setSpacing(4)
        y_row.addWidget(self.pixel_y_spin)
        y_row.addWidget(self.world_y_edit)
        y_widget = QWidget(detail_group)
        y_widget.setLayout(y_row)

        self.properties_form.addRow("World Frame", self.world_frame_combo)
        self.properties_form.addRow("X (pix)", x_widget)
        self.properties_form.addRow("Y (pix)", y_widget)

        self.label_field_label = QLabel("Label", self)
        self.label_edit = QLineEdit(self)
        self.label_edit.setPlaceholderText("Label")
        self.properties_form.addRow(self.label_field_label, self.label_edit)

        self.color_combo = QComboBox(self)
        self.color_combo.addItems(self.COLOR_OPTIONS)
        if "white" in self.COLOR_OPTIONS:
            self.color_combo.setCurrentText("white")

        self.opacity_spin = QDoubleSpinBox(self)
        self.opacity_spin.setRange(0.05, 1.0)
        self.opacity_spin.setSingleStep(0.05)
        self.opacity_spin.setValue(1.0)

        color_row = QHBoxLayout()
        color_row.setContentsMargins(0, 0, 0, 0)
        color_row.setSpacing(6)
        color_row.addWidget(self.color_combo)
        opacity_label = QLabel("Opacity", self)
        color_row.addWidget(opacity_label)
        color_row.addWidget(self.opacity_spin)
        color_container = QWidget(self)
        color_container.setLayout(color_row)
        self.properties_form.addRow("Color", color_container)

        # Symbol Controls
        self.symbol_controls = self._build_symbol_controls()
        size_label = QLabel("Size", detail_group)
        size_row_widget = QWidget(detail_group)
        size_row_layout = QHBoxLayout(size_row_widget)
        size_row_layout.setContentsMargins(0, 0, 0, 0)
        size_row_layout.setSpacing(6)
        size_row_layout.addWidget(self.symbol_controls.size_spin)
        size_row_layout.addSpacing(8)
        size_row_layout.addWidget(QLabel("Linewidth", size_row_widget))
        size_row_layout.addWidget(self.symbol_controls.linewidth_spin)
        size_row_layout.addStretch(1)
        self.properties_form.addRow(size_label, size_row_widget)
        self.properties_form.addRow(self.symbol_controls.edge_color_label, self.symbol_controls.edge_color_combo)
        self.property_rows["symbol"].extend([(size_label, size_row_widget), (self.symbol_controls.edge_color_label, self.symbol_controls.edge_color_combo)])

        # Line Controls
        self.line_controls = self._build_line_controls()
        length_label = QLabel("Length", detail_group)
        length_row_widget = QWidget(detail_group)
        length_row_layout = QHBoxLayout(length_row_widget)
        length_row_layout.setContentsMargins(0, 0, 0, 0)
        length_row_layout.setSpacing(6)
        length_row_layout.addWidget(self.line_controls.length_spin)
        length_row_layout.addWidget(self.line_controls.unit_combo)
        length_row_layout.addStretch(1)
        self.properties_form.addRow(length_label, length_row_widget)
        angle_label = QLabel("Angle (deg)", detail_group)
        angle_row_widget = QWidget(detail_group)
        angle_row_layout = QHBoxLayout(angle_row_widget)
        angle_row_layout.setContentsMargins(0, 0, 0, 0)
        angle_row_layout.setSpacing(6)
        angle_row_layout.addWidget(self.line_controls.angle_spin)
        angle_row_layout.addSpacing(8)
        angle_row_layout.addWidget(QLabel("Linewidth", angle_row_widget))
        angle_row_layout.addWidget(self.line_controls.linewidth_spin)
        angle_row_layout.addStretch(1)
        self.properties_form.addRow(angle_label, angle_row_widget)
        style_label = QLabel("Style", detail_group)
        self.properties_form.addRow(style_label, self.line_controls.style_combo)
        self.property_rows["line"].extend([(length_label, length_row_widget), (angle_label, angle_row_widget), (style_label, self.line_controls.style_combo)])

        # Text Controls
        self.text_controls = self._build_text_controls()
        font_size_label = QLabel("Font Size", detail_group)
        self.properties_form.addRow(font_size_label, self.text_controls.font_size_spin)
        rotation_label = QLabel("Angle (deg)", detail_group)
        self.properties_form.addRow(rotation_label, self.text_controls.rotation_spin)
        font_label = QLabel("Font", detail_group)
        self.properties_form.addRow(font_label, self.text_controls.font_combo)
        self.property_rows["text"].extend([(font_size_label, self.text_controls.font_size_spin), (rotation_label, self.text_controls.rotation_spin), (font_label, self.text_controls.font_combo)])

        # Set initial visibility
        self._set_property_rows_visible("symbol", "o")

        main_layout.addWidget(detail_group)

        list_group = QGroupBox("Markers", self)
        list_layout = QVBoxLayout(list_group)
        list_layout.setContentsMargins(8, 8, 8, 8)
        list_layout.setSpacing(6)

        self.marker_list = DeselectableListWidget(self)
        self.marker_list.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        list_layout.addWidget(self.marker_list)

        list_button_row = QHBoxLayout()
        list_button_row.setSpacing(8)
        self.delete_button = QPushButton("Delete", self)
        self.delete_all_button = QPushButton("Delete All", self)
        self.save_button = QPushButton("Save", self)
        self.load_button = QPushButton("Load", self)
        self.delete_button.setDefault(False)
        self.delete_button.setAutoDefault(False)
        self.delete_all_button.setDefault(False)
        self.delete_all_button.setAutoDefault(False)
        self.save_button.setDefault(False)
        self.save_button.setAutoDefault(False)
        self.load_button.setDefault(False)
        self.load_button.setAutoDefault(False)
        list_button_row.addWidget(self.save_button)
        list_button_row.addWidget(self.load_button)
        list_button_row.addWidget(self.delete_button)
        list_button_row.addWidget(self.delete_all_button)
        list_button_row.addStretch()
        list_layout.addLayout(list_button_row)

        main_layout.addWidget(list_group)

    def _build_symbol_controls(self) -> SymbolControls:
        size_spin = QDoubleSpinBox()
        size_spin.setRange(1.0, 100.0)
        size_spin.setSingleStep(1.0)
        size_spin.setValue(8.0)

        linewidth_spin = QDoubleSpinBox()
        linewidth_spin.setRange(0.0, 10.0)
        linewidth_spin.setSingleStep(0.1)
        linewidth_spin.setValue(0.0)

        edge_color_label = QLabel("Edge Color")
        edge_color_combo = QComboBox()
        edge_color_combo.addItems(self.COLOR_OPTIONS)

        return SymbolControls(
            widget=QWidget(),  # Dummy widget, no longer used for layout
            size_spin=size_spin,
            linewidth_spin=linewidth_spin,
            edge_color_label=edge_color_label,
            edge_color_combo=edge_color_combo,
        )

    def _build_line_controls(self) -> LineControls:
        length_spin = QDoubleSpinBox()
        length_spin.setRange(0.1, 10_000.0)
        length_spin.setSingleStep(1.0)
        length_spin.setValue(20.0)

        unit_combo = QComboBox()
        unit_combo.addItems(self.UNIT_OPTIONS)

        angle_spin = QDoubleSpinBox()
        angle_spin.setRange(-360.0, 360.0)
        angle_spin.setSingleStep(1.0)
        angle_spin.setValue(0.0)

        linewidth_spin = QDoubleSpinBox()
        linewidth_spin.setRange(0.1, 20.0)
        linewidth_spin.setSingleStep(0.2)
        linewidth_spin.setValue(1.0)

        style_combo = QComboBox()
        style_combo.addItem("Solid", "solid")
        style_combo.addItem("Dashed", "dashed")
        style_combo.addItem("Dotted", "dotted")
        style_combo.addItem("Arrow →", "arrow")
        style_combo.addItem("Arrow ↔", "double_arrow")
        style_combo.addItem("Scale", "scale")

        return LineControls(
            widget=QWidget(), # Dummy
            length_spin=length_spin,
            unit_combo=unit_combo,
            angle_spin=angle_spin,
            linewidth_spin=linewidth_spin,
            style_combo=style_combo,
        )

    def _build_text_controls(self) -> TextControls:
        font_size_spin = QDoubleSpinBox()
        font_size_spin.setRange(6.0, 72.0)
        font_size_spin.setSingleStep(1.0)
        font_size_spin.setValue(14.0)

        rotation_spin = QDoubleSpinBox()
        rotation_spin.setRange(-360.0, 360.0)
        rotation_spin.setSingleStep(1.0)
        rotation_spin.setValue(0.0)

        font_combo = QComboBox()
        for family in self.FONT_OPTIONS:
            font_combo.addItem(family)
        font_combo.setCurrentText("Arial")

        return TextControls(
            widget=QWidget(), # Dummy
            font_size_spin=font_size_spin,
            rotation_spin=rotation_spin,
            font_combo=font_combo,
        )

    def _set_property_rows_visible(self, kind: str, symbol: Optional[str] = None) -> None:
        for row_kind in self.property_rows:
            for label, widget in self.property_rows[row_kind]:
                is_visible = row_kind == kind
                label.setVisible(is_visible)
                widget.setVisible(is_visible)
        if kind == "symbol":
            symbol_value = symbol
            if symbol_value is None:
                index = self.type_combo.currentIndex()
                kind_data = self.TYPE_OPTIONS[index][1]
                symbol_value = kind_data.get("symbol")
            is_plus_cross = symbol_value in {"+", "x"}
            self.symbol_controls.edge_color_label.setVisible(not is_plus_cross)
            self.symbol_controls.edge_color_combo.setVisible(not is_plus_cross)
        else:
            self.symbol_controls.edge_color_label.setVisible(False)
            self.symbol_controls.edge_color_combo.setVisible(False)

    def _capture_detail_group_height(self, kind: str) -> None:
        group = self.detail_group
        if group is None:
            return
        height = int(group.sizeHint().height() or 0)
        if height <= 0:
            return
        previous = self._detail_group_heights.get(kind, 0)
        if height > previous:
            self._detail_group_heights[kind] = height
        max_height = max(self._detail_group_heights.values(), default=0)
        if max_height and max_height != self._detail_group_fixed_height:
            self._detail_group_fixed_height = max_height
            group.setMinimumHeight(max_height)
            group.setMaximumHeight(max_height)

    def _precompute_detail_group_heights(self) -> None:
        group = self.detail_group
        if group is None:
            return
        current_index = self.type_combo.currentIndex()
        current_data = self.TYPE_OPTIONS[current_index][1]
        original_kind = current_data.get("kind", "symbol")
        original_symbol = current_data.get("symbol")
        self._suspend_updates = True
        try:
            for _, data in self.TYPE_OPTIONS:
                kind = data.get("kind")
                if kind not in self.property_rows:
                    continue
                self._set_property_rows_visible(kind, data.get("symbol"))
                self._update_label_field_visibility(kind)
                group.adjustSize()
                self._capture_detail_group_height(kind)
        finally:
            self._set_property_rows_visible(original_kind, original_symbol)
            self._update_label_field_visibility(original_kind)
            self._suspend_updates = False
        self._capture_detail_group_height(original_kind)

    def _make_coordinate_spin(self, parent: QWidget) -> QDoubleSpinBox:
        spin = QDoubleSpinBox(parent)
        spin.setDecimals(3)
        spin.setRange(-1_000_000.0, 1_000_000.0)
        spin.setSingleStep(0.5)
        spin.setEnabled(False)
        return spin

    def _make_world_line_edit(self, parent: QWidget, placeholder: str) -> QLineEdit:
        edit = QLineEdit(parent)
        edit.setPlaceholderText(placeholder)
        edit.setEnabled(False)
        return edit

    def _populate_world_frame_combo(self) -> None:
        current = normalize_display_frame(self._get_shared_display_frame())
        self.world_frame_combo.blockSignals(True)
        self.world_frame_combo.clear()
        frames = available_display_frames(getattr(self.viewer, "wcs", None))
        for frame in frames:
            self.world_frame_combo.addItem(display_frame_label(frame), frame)
        idx = self.world_frame_combo.findData(current)
        if idx < 0:
            idx = self.world_frame_combo.findData("native")
        if idx >= 0:
            self.world_frame_combo.setCurrentIndex(idx)
        self.world_frame_combo.setEnabled(self.world_frame_combo.count() > 1)
        self.world_frame_combo.blockSignals(False)

    def _current_world_frame(self) -> str:
        data = self.world_frame_combo.currentData()
        frame = normalize_display_frame(data)
        if not frame_is_available(getattr(self.viewer, "wcs", None), frame):
            return "native"
        return frame

    def _shared_world_vector(self, viewer=None):
        target = viewer or self.viewer
        values = []
        for name in ("_get_shared_world_x", "_get_shared_world_y", "_get_shared_world_z", "_get_shared_world_s"):
            getter = getattr(target, name, None)
            if callable(getter):
                try:
                    values.append(float(getter()))
                except Exception:
                    values.append(None)
            else:
                values.append(None)
        return values

    def _get_shared_display_frame(self) -> str:
        getter = getattr(self.viewer, "_get_shared_display_frame", None)
        if callable(getter):
            try:
                return getter()
            except Exception:
                return "native"
        return "native"

    # ------------------------------------------------------------------
    # Signal wiring
    def _connect_signals(self) -> None:
        self.type_combo.currentIndexChanged.connect(self._on_type_changed)
        self.placement_toggle.toggled.connect(self._on_placement_toggled)
        if self.link_all_checkbox is not None:
            self.link_all_checkbox.toggled.connect(self._on_link_all_toggled)

        self.marker_list.itemSelectionChanged.connect(self._on_marker_list_selection)
        self.marker_list.delete_pressed.connect(self._on_delete_marker)
        self.delete_button.clicked.connect(self._on_delete_marker)
        self.delete_all_button.clicked.connect(self._on_delete_all)
        self.save_button.clicked.connect(self._on_save_json)
        self.load_button.clicked.connect(self._on_load_json)

        self.label_edit.editingFinished.connect(self._on_label_changed)
        self.label_edit.returnPressed.connect(self._on_label_changed)
        self.color_combo.currentTextChanged.connect(lambda _: self._apply_common_style())
        self.opacity_spin.valueChanged.connect(lambda _: self._apply_common_style())

        self.pixel_x_spin.valueChanged.connect(self._on_pixel_spin_changed)
        self.pixel_y_spin.valueChanged.connect(self._on_pixel_spin_changed)
        self.world_x_edit.editingFinished.connect(self._on_world_edit_finished)
        self.world_y_edit.editingFinished.connect(self._on_world_edit_finished)
        self.world_frame_combo.currentIndexChanged.connect(self._on_world_frame_changed)

        self.symbol_controls.size_spin.valueChanged.connect(lambda _: self._apply_symbol_style())
        self.symbol_controls.linewidth_spin.valueChanged.connect(lambda _: self._apply_symbol_style())
        self.symbol_controls.edge_color_combo.currentTextChanged.connect(
            lambda _: self._apply_symbol_style()
        )

        self.line_controls.length_spin.valueChanged.connect(lambda _: self._apply_line_style())
        self.line_controls.unit_combo.currentTextChanged.connect(self._on_line_unit_changed)
        self.line_controls.angle_spin.valueChanged.connect(lambda _: self._apply_line_style())
        self.line_controls.linewidth_spin.valueChanged.connect(lambda _: self._apply_line_style())
        self.line_controls.style_combo.currentIndexChanged.connect(self._on_line_style_changed)

        self.text_controls.font_size_spin.valueChanged.connect(lambda _: self._apply_text_style())
        self.text_controls.rotation_spin.valueChanged.connect(lambda _: self._apply_text_style())
        self.text_controls.font_combo.currentTextChanged.connect(lambda _: self._apply_text_style())

        self.marker_manager.markers_changed.connect(self._on_markers_changed)
        self.marker_manager.selection_changed.connect(self._on_manager_selection_changed)
        self.marker_manager.geometry_changed.connect(self._on_marker_geometry_changed)

    # ------------------------------------------------------------------
    # Marker list and selection helpers
    def _refresh_marker_list(self) -> None:
        self._suspend_selection_sync = True
        try:
            selected_lookup = getattr(self.marker_manager, "selected_marker", None)
            selected_marker = selected_lookup() if callable(selected_lookup) else None
            selected_ids = set(self._selected_marker_ids())
            current_id = None
            if selected_marker is not None:
                current_id = selected_marker.marker_id
            elif self._current_marker_id:
                current_id = self._current_marker_id
            if current_id:
                selected_ids.add(current_id)
            self.marker_list.clear()
            items: List[Tuple[str, Marker]] = []
            for plane, layer in self.marker_manager._layers.items():
                for marker in layer.markers.values():
                    items.append((plane, marker))
            # Stable order: plane order xy -> xz -> zy, then insertion order
            plane_order = {"xy": 0, "xz": 1, "zy": 2}
            items.sort(key=lambda entry: plane_order.get(entry[0], 99))
            for plane, marker in items:
                display_text = self._list_entry_text(marker)
                item = QListWidgetItem(display_text, self.marker_list)
                item.setData(Qt.ItemDataRole.UserRole, marker.marker_id)
                self.marker_list.addItem(item)
                if marker.marker_id in selected_ids:
                    item.setSelected(True)
            if selected_ids:
                primary = self._primary_marker()
                if primary is not None:
                    self._current_marker_id = primary.marker_id
                    self._populate_detail_fields(primary, preserve_selection=True)
                else:
                    self._clear_detail_fields()
            else:
                self._clear_detail_fields()
        finally:
            self._suspend_selection_sync = False

    def _selected_marker_ids(self) -> List[str]:
        return [
            item.data(Qt.ItemDataRole.UserRole)
            for item in self.marker_list.selectedItems()
            if item.data(Qt.ItemDataRole.UserRole)
        ]

    def _selected_markers(self) -> List[Marker]:
        markers: List[Marker] = []
        for marker_id in self._selected_marker_ids():
            marker = self.marker_manager.marker_for_id(marker_id)
            if marker is not None:
                markers.append(marker)
        return markers

    def _primary_marker(self) -> Optional[Marker]:
        markers = self._selected_markers()
        if markers:
            return markers[0]
        if self._current_marker_id is None:
            return None
        return self.marker_manager.marker_for_id(self._current_marker_id)

    def _current_marker(self) -> Optional[Marker]:
        return self._primary_marker()

    def set_selection_by_id(self, marker_ids: List[str]) -> None:
        self._suspend_selection_sync = True
        self.marker_list.clearSelection()
        ids_to_select = set(marker_ids)
        for i in range(self.marker_list.count()):
            item = self.marker_list.item(i)
            if item.data(Qt.ItemDataRole.UserRole) in ids_to_select:
                item.setSelected(True)
        self._suspend_selection_sync = False
        self._on_marker_list_selection()

    def _list_entry_text(self, marker: Marker) -> str:
        kind_desc = self._kind_display(marker)
        color_value = marker.style.color
        
        # For Plus/Cross markers (SymbolMarker), the primary color is stored in edgecolor.
        # If color is 'none' or missing, check edgecolor.
        if not color_value or (isinstance(color_value, str) and color_value.lower() == "none"):
            if isinstance(marker, SymbolMarker):
                edge = marker.style.edgecolor
                if edge and str(edge).lower() != "none":
                    color_value = edge
        
        if not color_value:
             color_value = "auto"
        elif isinstance(color_value, str) and color_value.lower() == "none":
             color_value = "none"
        color_text = str(color_value)

        label = (marker.label or "").strip()
        detail = ""
        if isinstance(marker, TextMarker):
            content = (getattr(marker, "text", "") or label).strip()
            if content:
                detail = f' "{content}"'
        elif label:
            detail = f' "{label}"'

        plane = marker.plane.upper()
        viewer = self._viewer_for_marker(marker)
        converter = getattr(viewer, "converter", None) if viewer else None
        wcs = getattr(converter, "wcs", None) if converter is not None else getattr(viewer, "wcs", None)
        axis_indices = self._plane_axis_indices(marker.plane, wcs)
        world_values = self._world_strings_from_marker(
            marker,
            converter=converter,
            axis_indices=axis_indices,
            viewer=viewer,
        )
        if world_values and any(world_values):
            world_text = ", ".join(world_values)
        else:
            world_text = "--"
        return f"{kind_desc}{detail} ({color_text}) [{plane}] {world_text}"

    def _kind_display(self, marker: Marker) -> str:
        if isinstance(marker, SymbolMarker):
            symbol = getattr(marker, "symbol", "o")
            symbol_map = {
                "o": "Circle",
                "s": "Square",
                "D": "Diamond",
                "d": "Diamond",
                "^": "Triangle Up",
                "v": "Triangle Down",
                "*": "Star",
                "+": "Plus",
                "x": "Cross",
            }
            return symbol_map.get(symbol, f"Symbol ({symbol})")
        if isinstance(marker, LineMarker):
            return "Line"
        if isinstance(marker, TextMarker):
            return "Text"
        return marker.kind.capitalize()

    # ------------------------------------------------------------------
    # Selection and detail updates
    def _clear_detail_fields(self) -> None:
        self._suspend_updates = True
        try:
            self._current_marker_id = None
            self._pending_selection_id = None
            self.label_edit.clear()
            if "white" in self.COLOR_OPTIONS:
                self.color_combo.setCurrentText("white")
            else:
                self.color_combo.setCurrentIndex(0)
            self.opacity_spin.setValue(1.0)

            self._set_position_controls_enabled(False)
            self.pixel_x_spin.blockSignals(True)
            self.pixel_y_spin.blockSignals(True)
            self.pixel_x_spin.setValue(0.0)
            self.pixel_y_spin.setValue(0.0)
            self.pixel_x_spin.blockSignals(False)
            self.pixel_y_spin.blockSignals(False)
            self.world_x_edit.blockSignals(True)
            self.world_y_edit.blockSignals(True)
            self.world_x_edit.clear()
            self.world_y_edit.clear()
            self.world_x_edit.blockSignals(False)
            self.world_y_edit.blockSignals(False)
            self.line_controls.style_combo.setCurrentIndex(0)
            current_kind = self.TYPE_OPTIONS[self.type_combo.currentIndex()][1].get("kind", "symbol")
            self._update_label_field_visibility(current_kind)
            if current_kind == "text":
                template_text = self._last_text_value or "Text"
                self.label_edit.setText(template_text)
            fallback_font = next((font for font in self.FONT_OPTIONS if font != "STIXGeneral"), self.FONT_OPTIONS[0])
            self.text_controls.font_combo.setCurrentText(fallback_font)
            self.pixel_x_spin.setEnabled(True)
            self.pixel_y_spin.setEnabled(True)
            self.world_x_edit.setEnabled(True)
            self.world_y_edit.setEnabled(True)
        finally:
            self._suspend_updates = False

    def _populate_detail_fields(self, marker: Marker, *, preserve_selection: bool = False) -> None:
        self._suspend_updates = True
        focus_label_field = False
        fallback_font_option = next((font for font in self.FONT_OPTIONS if font != "STIXGeneral"), self.FONT_OPTIONS[0])
        try:
            self._current_plane = marker.plane
            self._current_marker_id = marker.marker_id
            if hasattr(self.marker_manager, "set_active_plane"):
                current_plane_getter = getattr(self.marker_manager, "active_plane", None)
                active_plane = current_plane_getter() if callable(current_plane_getter) else None
                if active_plane != marker.plane:
                    self._pending_selection_id = marker.marker_id
                    self.marker_manager.set_active_plane(marker.plane)
            selected_markers = self._selected_markers()
            multi_selected = len(selected_markers) > 1
            allow_multi_position = multi_selected and self._viewer_supports_position_edit(selected_markers)
            self._set_position_controls_enabled(not multi_selected or allow_multi_position)
            style = marker.style
            type_index = next(
                (i for i, (label, data) in enumerate(self.TYPE_OPTIONS) if self._type_matches_marker(data, marker)),
                0,
            )
            if self.type_combo.currentIndex() != type_index:
                self.type_combo.setCurrentIndex(type_index)
            kind = self.TYPE_OPTIONS[type_index][1].get("kind", "symbol")
            self._update_label_field_visibility(kind)
            if kind == "text":
                self.label_edit.setText(marker.text)
            else:
                self.label_edit.setText(marker.label)
            color_value = style.color
            if isinstance(color_value, str) and color_value.lower() == "none":
                self.color_combo.setCurrentText("None")
            elif isinstance(color_value, str) and color_value in self.COLOR_OPTIONS:
                self.color_combo.setCurrentText(color_value)
            elif "white" in self.COLOR_OPTIONS:
                self.color_combo.setCurrentText("white")
            else:
                self.color_combo.setCurrentIndex(0)
            self.opacity_spin.setValue(style.opacity)

            if isinstance(marker, SymbolMarker):
                
                is_plus_cross = getattr(marker, "symbol", None) in {"+", "x"}
                self.symbol_controls.edge_color_label.setVisible(not is_plus_cross)
                self.symbol_controls.edge_color_combo.setVisible(not is_plus_cross)

                self.symbol_controls.size_spin.setValue(style.size)
                linewidth_value = style.linewidth or 0.0
                if is_plus_cross:
                    linewidth_value = max(1.0, linewidth_value)
                self.symbol_controls.linewidth_spin.setValue(linewidth_value)

                if is_plus_cross:
                    edge_color = style.edgecolor or style.color
                    if (
                        isinstance(edge_color, str)
                        and edge_color in self.COLOR_OPTIONS
                    ):
                        self.color_combo.setCurrentText(edge_color)
                    else:
                        self.color_combo.setCurrentText("white")
                else:
                    edge_color_value = style.edgecolor
                    if (
                        isinstance(edge_color_value, str)
                        and edge_color_value.lower() == "none"
                    ):
                        self.symbol_controls.edge_color_combo.setCurrentText("None")
                    elif (
                        isinstance(edge_color_value, str)
                        and edge_color_value in self.COLOR_OPTIONS
                    ):
                        self.symbol_controls.edge_color_combo.setCurrentText(
                            edge_color_value
                        )
                    else:
                        self.symbol_controls.edge_color_combo.setCurrentText("None")

            elif isinstance(marker, LineMarker):
                #self.detail_stack.setCurrentIndex(1)
                self._update_line_detail_fields(marker, style=style)

            elif isinstance(marker, TextMarker):
                fallback_font_option = next((font for font in self.FONT_OPTIONS if font != "STIXGeneral"), self.FONT_OPTIONS[0])
                self.text_controls.font_size_spin.setValue(style.size)
                self.text_controls.rotation_spin.setValue(style.rotation)
                text_value = marker.text or ""
                if not text_value:
                    text_value = "Text"
                    marker.update_text(text_value)
                self.label_edit.setText(text_value)
                self._last_text_value = text_value
                if "$" in text_value:
                    self.text_controls.font_combo.setCurrentText("STIXGeneral")
                else:
                    family = style.font_family or fallback_font_option
                    if family not in self.FONT_OPTIONS:
                        self.text_controls.font_combo.addItem(family)
                    self.text_controls.font_combo.setCurrentText(family)
                if not text_value.strip():
                    focus_label_field = True
            self._update_position_fields(marker)
        finally:
            self._suspend_updates = False
        self._ensure_marker_selected(marker, preserve_existing=preserve_selection)
        if focus_label_field:
            def _focus_text_field() -> None:
                self.label_edit.setFocus(Qt.FocusReason.OtherFocusReason)
                self.label_edit.selectAll()
            QTimer.singleShot(0, _focus_text_field)

    def _type_matches_marker(self, data: Dict[str, str], marker: Marker) -> bool:
        if data.get("kind") == "symbol" and isinstance(marker, SymbolMarker):
            symbol = data.get("symbol", "")
            return symbol == getattr(marker, "symbol", None)
        if data.get("kind") == "line" and isinstance(marker, LineMarker):
            return True
        if data.get("kind") == "text" and isinstance(marker, TextMarker):
            return True
        return False

    def _set_position_controls_enabled(self, enabled: bool) -> None:
        for widget in (self.pixel_x_spin, self.pixel_y_spin):
            widget.setEnabled(enabled)
        for widget in (self.world_x_edit, self.world_y_edit):
            widget.setEnabled(enabled)

    def _normalized_color_value(self, value: Optional[str]) -> Optional[str]:
        if value is None:
            return None
        lowered = value.lower() if isinstance(value, str) else value
        if lowered == "none":
            return "none"
        return value

    def _current_kind(self) -> str:
        return self.TYPE_OPTIONS[self.type_combo.currentIndex()][1].get("kind", "symbol")

    def _style_configuration_for_kind(self, kind: str) -> Dict[str, object]:
        style_dict: Dict[str, object] = {
            "color": self._normalized_color_value(self.color_combo.currentText()),
            "opacity": self.opacity_spin.value(),
        }
        kwargs: Dict[str, object] = {}
        style_dict["edgecolor"] = None

        if kind == "symbol":
            size = self.symbol_controls.size_spin.value()
            style_dict["size"] = size
            symbol = self.TYPE_OPTIONS[self.type_combo.currentIndex()][1].get("symbol", "o")
            linewidth_value = self.symbol_controls.linewidth_spin.value()
            if symbol in {"+", "x"}:
                linewidth_value = max(1.0, linewidth_value)
                style_dict["color"] = "none"
                style_dict["edgecolor"] = self._normalized_color_value(
                    self.color_combo.currentText()
                )
            else:
                style_dict["edgecolor"] = self._normalized_color_value(
                    self.symbol_controls.edge_color_combo.currentText()
                )
            style_dict["linewidth"] = linewidth_value
            style_dict["marker_symbol"] = symbol
            kwargs["symbol"] = symbol
        elif kind == "line":
            size = self.symbol_controls.size_spin.value()
            style_dict["size"] = size
            style_dict["linewidth"] = self.line_controls.linewidth_spin.value()
            raw_length = self.line_controls.length_spin.value()
            unit = self.line_controls.unit_combo.currentText()
            pixels = self._convert_length_to_pixels(raw_length, unit, plane=self._current_plane)
            kwargs.update(
                {
                    "length": pixels,
                    "angle_deg": self.line_controls.angle_spin.value(),
                    "unit": unit,
                    "length_source_value": raw_length,
                    "style_mode": self.line_controls.style_combo.currentData(),
                }
            )
        elif kind == "text":
            style_dict["size"] = self.text_controls.font_size_spin.value()
            style_dict["rotation"] = self.text_controls.rotation_spin.value()
            text_value = self.label_edit.text().strip()
            kwargs["text"] = text_value
            if "$" in text_value:
                style_dict["font_family"] = "STIXGeneral"
            else:
                style_dict["font_family"] = self.text_controls.font_combo.currentText()
        else:
            style_dict["size"] = self.symbol_controls.size_spin.value()

        style = MarkerStyle.from_dict(style_dict)
        kwargs["style"] = style

        label = self.label_edit.text().strip()
        if kind != "text":
            if self._suppress_label_from_text and label == self._last_text_value:
                label = ""
            if label:
                kwargs["label"] = label
        return kwargs

    def _resolve_current_plane(self) -> str:
        current_plane = self._current_plane
        selected_lookup = getattr(self.marker_manager, "selected_marker", None)
        selected_marker = selected_lookup() if callable(selected_lookup) else None
        active_plane_getter = getattr(self.marker_manager, "active_plane", None)
        manager_plane = active_plane_getter() if callable(active_plane_getter) else None
        if selected_marker is not None:
            current_plane = selected_marker.plane
        elif manager_plane:
            current_plane = manager_plane
        if not current_plane:
            current_plane = getattr(self.viewer, "plane", "xy")
        self._current_plane = current_plane
        return current_plane

    def _world_inputs_to_pixels(
        self,
        plane: str,
        world_inputs: Tuple[str, str],
    ) -> Optional[Tuple[float, float]]:
        viewer = self._viewer_for_plane(plane)
        converter = getattr(viewer, "converter", None) if viewer else None
        wcs = getattr(converter, "wcs", None) if converter else None
        axis_indices = self._plane_axis_indices(plane, wcs)
        if converter is None or wcs is None or axis_indices is None:
            return None
        cleaned_inputs = [value.strip() for value in world_inputs]
        if not all(cleaned_inputs):
            return None
        try:
            ref_pixel = getattr(wcs, "wcs", None).crpix if hasattr(wcs, "wcs") else wcs.crpix
            ref_world = wcs.wcs_pix2world([ref_pixel], 0)[0]
        except Exception:
            return None
        target_world = list(ref_world)
        fallback_world = self._shared_world_vector(viewer)
        native_values = plane_inputs_to_native(
            wcs,
            plane,
            cleaned_inputs[0],
            cleaned_inputs[1],
            frame=self._current_world_frame(),
            fallback_native_world=fallback_world,
        )
        if native_values is None:
            return None
        for value, idx in zip(native_values, axis_indices):
            target_world[idx] = value
        try:
            pix_coords = converter.world_to_pix(*target_world)
            x_pix = float(pix_coords[axis_indices[0]])
            y_pix = float(pix_coords[axis_indices[1]])
        except Exception:
            return None
        return (x_pix, y_pix)

    def _create_marker_from_coordinates(self, plane: str, x: float, y: float) -> Optional[Marker]:
        entry = self.TYPE_OPTIONS[self.type_combo.currentIndex()]
        kind = entry[1].get("kind", "symbol")
        kwargs = dict(self._style_configuration_for_kind(kind))
        marker: Optional[Marker] = None
        if kind == "symbol":
            marker = self.marker_manager.create_symbol_marker(plane, (x, y), **kwargs)
        elif kind == "line":
            marker = self.marker_manager.create_line_marker(plane, (x, y), **kwargs)
        elif kind == "text":
            marker = self.marker_manager.create_text_marker(plane, (x, y), **kwargs)
        else:
            marker = self.marker_manager.create_symbol_marker(plane, (x, y), **kwargs)
        if marker is None:
            return None
        if kind == "text" and self.placement_toggle.isChecked():
            self.marker_manager.select_marker(None)
            self._current_marker_id = None
            self._on_deselect_markers()
            self._refresh_marker_list()
            self._focus_label_edit()
        else:
            self.marker_manager.select_marker(marker)
            self._refresh_marker_list()
            self._populate_detail_fields(marker)
        return marker

    def _focus_label_edit(self) -> None:
        def _focus() -> None:
            try:
                if self.isMinimized():
                    self.showNormal()
                self.raise_()
                self.activateWindow()
            except Exception:
                pass
            self.label_edit.setFocus(Qt.FocusReason.ActiveWindowFocusReason)
            self.label_edit.selectAll()
        for delay_ms in (0, 40, 120):
            QTimer.singleShot(delay_ms, _focus)

    def focus_text_entry_for_marker(self, marker: Optional[Marker] = None) -> None:
        """Focus the label editor when the target marker is a text marker."""
        target = marker if marker is not None else self._current_marker()
        if not isinstance(target, TextMarker):
            return
        if not self.isVisible():
            return
        self._focus_label_edit()

    def _update_label_field_visibility(self, kind: str) -> None:
        if kind == "text":
            self.label_field_label.setText("Text")
            self.label_edit.setPlaceholderText("Text")
        else:
            self.label_field_label.setText("Label")
            self.label_edit.setPlaceholderText("Label")
        self.label_field_label.setVisible(True)
        self.label_edit.setVisible(True)

    def _update_position_fields(self, marker: Marker) -> None:
        pixel = list(marker.pixel)
        if len(pixel) < 2:
            pixel.extend([0.0] * (2 - len(pixel)))
        x_pix, y_pix = float(pixel[0]), float(pixel[1])
        self.pixel_x_spin.blockSignals(True)
        self.pixel_y_spin.blockSignals(True)
        self.pixel_x_spin.setValue(x_pix)
        self.pixel_y_spin.setValue(y_pix)
        self.pixel_x_spin.blockSignals(False)
        self.pixel_y_spin.blockSignals(False)

        previous_world_x = self.world_x_edit.text()
        previous_world_y = self.world_y_edit.text()

        viewer = self._viewer_for_plane(marker.plane)
        converter = getattr(viewer, "converter", None) if viewer else None
        wcs = getattr(converter, "wcs", None) if converter else None
        axis_indices = self._plane_axis_indices(marker.plane, wcs)

        world_editable = bool(converter is not None and wcs is not None and axis_indices is not None)
        world_texts: Optional[Tuple[str, str]] = None

        if world_editable:
            try:
                pixel_vector = self._build_pixel_vector(marker.plane, x_pix, y_pix, wcs)
                native_world = wcs.wcs_pix2world([pixel_vector], 0)[0]
                frame = self._current_world_frame()
                fallback_world = self._shared_world_vector(viewer)
                wx_val, wy_val, wx_axis, wy_axis = plane_values_for_display(
                    wcs,
                    marker.plane,
                    native_world[axis_indices[0]],
                    native_world[axis_indices[1]],
                    frame=frame,
                    fallback_native_world=fallback_world,
                )
                if converter is not None and hasattr(converter, "format_world_coordinate"):
                    wx_text = converter.format_world_coordinate(wx_val, wx_axis)
                    wy_text = converter.format_world_coordinate(wy_val, wy_axis)
                else:
                    wx_text = str(wx_val)
                    wy_text = str(wy_val)
                world_texts = (str(wx_text), str(wy_text))
            except Exception:
                world_texts = None

        if world_texts is None:
            world_texts = self._world_strings_from_marker(
                marker,
                converter=converter,
                axis_indices=axis_indices,
                viewer=viewer,
            )

        if not any(world_texts) and (previous_world_x or previous_world_y):
            world_texts = (previous_world_x, previous_world_y)

        self._set_world_fields(world_texts[0], world_texts[1], editable=world_editable)

    def _set_world_fields(self, x_text: str, y_text: str, *, editable: bool) -> None:
        self.world_x_edit.blockSignals(True)
        self.world_y_edit.blockSignals(True)
        self.world_x_edit.setText(x_text)
        self.world_y_edit.setText(y_text)
        self.world_x_edit.blockSignals(False)
        self.world_y_edit.blockSignals(False)
        self.world_x_edit.setEnabled(editable)
        self.world_y_edit.setEnabled(editable)

    def _world_strings_from_marker(
        self,
        marker: Marker,
        *,
        converter=None,
        axis_indices: Optional[Tuple[int, int]] = None,
        viewer=None,
    ) -> Tuple[str, str]:
        world = getattr(marker, "world", None)
        if not world or len(world) < 2:
            return ("", "")
        wcs = getattr(converter, "wcs", None) if converter is not None else None
        frame = self._current_world_frame()
        if wcs is not None and axis_indices:
            try:
                fallback_world = self._shared_world_vector(viewer)
                wx_val, wy_val, wx_axis, wy_axis = plane_values_for_display(
                    wcs,
                    marker.plane,
                    world[0],
                    world[1],
                    frame=frame,
                    fallback_native_world=fallback_world,
                )
                if converter is not None and hasattr(converter, "format_world_coordinate"):
                    return (
                        str(converter.format_world_coordinate(wx_val, wx_axis)),
                        str(converter.format_world_coordinate(wy_val, wy_axis)),
                    )
                return (str(wx_val), str(wy_val))
            except Exception:
                pass
        default_decimals = self._default_world_decimals(converter)
        if (
            converter is not None
            and hasattr(converter, "format_world_coordinate")
            and hasattr(converter, "get_axis_types")
            and axis_indices
        ):
            try:
                axis_types = converter.get_axis_types()
            except Exception:
                axis_types = None
            if axis_types:
                formatted_values: List[str] = []
                for value, axis_idx in zip(world, axis_indices):
                    axis_type = None
                    if axis_idx is not None and 0 <= axis_idx < len(axis_types):
                        try:
                            axis_type = display_axis_type(axis_types[axis_idx], frame)
                            formatted = converter.format_world_coordinate(value, axis_type)
                        except Exception:
                            formatted = None
                        if formatted is not None:
                            formatted_values.append(str(formatted))
                            continue
                    decimals = self._auto_world_decimals_for_axis(
                        converter=converter,
                        wcs=wcs,
                        axis_index=axis_idx,
                        axis_type=axis_type,
                        fallback=default_decimals,
                    )
                    formatted_values.append(self._format_world_numeric_fallback(value, decimals))
                if len(formatted_values) == 2:
                    return (formatted_values[0], formatted_values[1])
        if axis_indices and len(axis_indices) == 2:
            x_decimals = self._auto_world_decimals_for_axis(
                converter=converter,
                wcs=wcs,
                axis_index=axis_indices[0],
                fallback=default_decimals,
            )
            y_decimals = self._auto_world_decimals_for_axis(
                converter=converter,
                wcs=wcs,
                axis_index=axis_indices[1],
                fallback=default_decimals,
            )
        else:
            x_decimals = default_decimals
            y_decimals = default_decimals
        return (
            self._format_world_numeric_fallback(world[0], x_decimals),
            self._format_world_numeric_fallback(world[1], y_decimals),
        )

    @staticmethod
    def _format_world_numeric_fallback(value, decimals: int) -> str:
        try:
            return f"{float(value):.{max(0, int(decimals))}f}"
        except Exception:
            return str(value)

    @staticmethod
    def _default_world_decimals(converter, fallback: int = 6) -> int:
        decimals = fallback
        config = getattr(converter, "config", None) if converter is not None else None
        if config is not None:
            try:
                decimals = int(config.get("number_decimals", decimals))
            except Exception:
                pass
        return max(0, int(decimals))

    @staticmethod
    def _auto_world_decimals_for_axis(
        *,
        converter,
        wcs,
        axis_index: Optional[int],
        axis_type: Optional[str] = None,
        fallback: int = 6,
        max_decimals: int = 10,
    ) -> int:
        try:
            base = max(0, int(fallback))
        except Exception:
            base = 6

        auto_enabled = True
        config = getattr(converter, "config", None) if converter is not None else None
        if isinstance(config, dict):
            auto_enabled = bool(config.get("auto_precision_digits", True))
        if not auto_enabled:
            return base

        if converter is not None and axis_type and hasattr(converter, "_effective_decimals"):
            try:
                return max(0, int(converter._effective_decimals(axis_type)))
            except Exception:
                pass

        if wcs is None or axis_index is None:
            return base
        try:
            step = abs(float(wcs.wcs.cdelt[int(axis_index)]))
        except Exception:
            return base
        if not math.isfinite(step) or step <= 0:
            return base
        try:
            decimals = int(math.ceil(-math.log10(step / 10.0)))
        except Exception:
            return base
        return max(0, min(int(max_decimals), decimals))

    def _update_line_detail_fields(self, marker: LineMarker, *, style: Optional[MarkerStyle] = None) -> None:
        style = style or marker.style
        unit_combo = self.line_controls.unit_combo
        target_unit = marker.unit if marker.unit in self.UNIT_OPTIONS else unit_combo.currentText()
        if marker.unit in self.UNIT_OPTIONS:
            unit_combo.blockSignals(True)
            index = unit_combo.findText(marker.unit, Qt.MatchFlag.MatchFixedString)
            if index >= 0:
                unit_combo.setCurrentIndex(index)
            unit_combo.blockSignals(False)
        else:
            target_unit = unit_combo.currentText()

        display_unit = unit_combo.currentText() or target_unit or "pixel"
        display_length = self._convert_length_from_pixels(marker.length, display_unit, plane=marker.plane)

        length_spin = self.line_controls.length_spin
        length_spin.blockSignals(True)
        length_spin.setValue(display_length)
        length_spin.blockSignals(False)

        angle_spin = self.line_controls.angle_spin
        angle_spin.blockSignals(True)
        angle_spin.setValue(marker.angle_deg)
        angle_spin.blockSignals(False)

        linewidth_spin = self.line_controls.linewidth_spin
        linewidth_spin.blockSignals(True)
        linewidth_spin.setValue(style.linewidth or 1.0)
        linewidth_spin.blockSignals(False)

        style_combo = self.line_controls.style_combo
        style_combo.blockSignals(True)
        found = False
        for idx in range(style_combo.count()):
            if style_combo.itemData(idx) == marker.style_mode:
                style_combo.setCurrentIndex(idx)
                found = True
                break
        if not found and style_combo.count():
            style_combo.setCurrentIndex(0)
        style_combo.blockSignals(False)

    def _update_text_detail_fields(self, marker: TextMarker) -> None:
        rotation_spin = self.text_controls.rotation_spin
        rotation_spin.blockSignals(True)
        rotation_spin.setValue(marker.style.rotation)
        rotation_spin.blockSignals(False)

    def _ensure_marker_selected(
        self,
        marker: Optional[Marker],
        *,
        sync_manager: bool = True,
        preserve_existing: bool = False,
    ) -> None:
        if marker is None:
            return
        if sync_manager:
            self._pending_selection_id = marker.marker_id
        if preserve_existing:
            selected_ids = set(self._selected_marker_ids())
        else:
            selected_ids = set()
        selected_ids.add(marker.marker_id)
        previous_flag = self._suspend_selection_sync
        self._suspend_selection_sync = True
        try:
            for idx in range(self.marker_list.count()):
                item = self.marker_list.item(idx)
                marker_id = item.data(Qt.ItemDataRole.UserRole)
                item.setSelected(marker_id in selected_ids)
        finally:
            self._suspend_selection_sync = previous_flag
        self._current_marker_id = marker.marker_id
        if sync_manager and hasattr(self.marker_manager, "select_marker"):
            try:
                self.marker_manager.select_marker(marker)
            except Exception:
                pass
        if not sync_manager:
            self._pending_selection_id = None

    def _plane_axis_indices(self, plane: Optional[str], wcs) -> Optional[Tuple[int, int]]:
        if wcs is None:
            return None
        fallback = plane_axis_indices(plane, wcs)
        viewer = self._viewer_for_plane(plane)
        if viewer is not None:
            resolver = getattr(viewer, "marker_axis_indices", None)
            if callable(resolver):
                try:
                    indices = resolver(plane)
                except Exception:
                    indices = None
                if indices:
                    try:
                        first, second = indices[:2]
                        return (int(first), int(second))
                    except Exception:
                        pass
        return fallback

    def _schedule_pending_selection_restore(self) -> None:
        pending_id = self._pending_selection_id
        if not pending_id:
            return
        marker = self.marker_manager.marker_for_id(pending_id)
        if marker is None:
            return
        def _restore():
            self._ensure_marker_selected(marker, preserve_existing=True)
        QTimer.singleShot(0, _restore)

    def _viewer_for_marker(self, marker: Marker):
        return self._viewer_for_plane(marker.plane)

    def _viewer_supports_link_all(self) -> bool:
        support_fn = getattr(self.viewer, "supports_marker_link_all", None)
        if callable(support_fn):
            try:
                return bool(support_fn())
            except Exception:
                return False
        return False

    def _viewer_link_all_enabled(self) -> bool:
        getter = getattr(self.viewer, "marker_link_all_enabled", None)
        if callable(getter):
            try:
                return bool(getter())
            except Exception:
                return False
        return False

    def _set_viewer_link_all(self, enabled: bool) -> None:
        setter = getattr(self.viewer, "set_marker_link_all", None)
        if callable(setter):
            try:
                setter(bool(enabled))
            except Exception:
                pass

    def _on_link_all_toggled(self, checked: bool) -> None:
        self._set_viewer_link_all(checked)

    def _viewer_supports_position_edit(self, markers: Iterable[Marker]) -> bool:
        markers = list(markers)
        if not markers:
            return False
        viewers = {self._viewer_for_marker(marker) for marker in markers}
        if None in viewers or len(viewers) != 1:
            return False
        viewer = viewers.pop()
        checker = getattr(viewer, "can_update_marker_positions", None)
        if callable(checker):
            try:
                return bool(checker(markers))
            except Exception:
                return False
        return False

    def _base_plane(self, plane: Optional[str]) -> str:
        plane_value = plane or "xy"
        viewer = None
        try:
            viewer = self._viewer_for_plane(plane_value)
        except Exception:
            viewer = None
        if viewer is not None:
            resolver = getattr(viewer, "marker_plane_base", None)
            if callable(resolver):
                try:
                    base = resolver(plane_value)
                    if base:
                        return str(base).lower()
                except Exception:
                    pass
        plane_key = str(plane_value).lower()
        for candidate in ("xy", "xz", "zy"):
            if candidate in plane_key:
                return candidate
        return plane_key

    def _get_display_slices(self, viewer) -> Optional[List[str]]:
        if viewer is None:
            return None
        displaymap = getattr(viewer, "displaymap", None)
        if displaymap is not None:
            slices = getattr(displaymap, "slices", None)
            if slices:
                return slices
        for attr in ("integ_slice", "projection_slices", "slice"):
            slices = getattr(viewer, attr, None)
            if slices and not callable(slices):
                return slices
        format_pix = getattr(viewer, "format_pix", None)
        if format_pix is not None:
            return getattr(format_pix, "slices", None)
        return None

    def _build_pixel_vector(self, plane: str, x_pix: float, y_pix: float, wcs) -> List[float]:
        naxis = getattr(wcs, "naxis", 0) or 0
        plane_key = self._base_plane(plane)
        x_fixed = float(self.viewer._get_shared_xpix())
        y_fixed = float(self.viewer._get_shared_ypix())
        z_fixed = float(self.viewer._get_shared_zpix())
        stokes_fixed = float(self.viewer._get_shared_spix())
        vector: List[float] = []
        for axis in range(naxis):
            if axis == 0:
                if plane_key == "zy":
                    vector.append(x_fixed)
                else:
                    vector.append(float(x_pix))
            elif axis == 1:
                if plane_key == "xz":
                    vector.append(y_fixed)
                else:
                    vector.append(float(y_pix))
            elif axis == 2:
                if plane_key == "xy":
                    vector.append(z_fixed)
                elif plane_key == "xz":
                    vector.append(float(y_pix))
                elif plane_key == "zy":
                    vector.append(float(x_pix))
                else:
                    vector.append(z_fixed)
            elif axis == 3:
                vector.append(stokes_fixed)
            else:
                vector.append(0.0)
        return vector

    # ------------------------------------------------------------------
    # Signal handlers
    def _on_marker_list_selection(self) -> None:
        if self._suspend_selection_sync:
            return
        markers = self._selected_markers()
        if not markers:
            if self._pending_selection_id:
                self._schedule_pending_selection_restore()
                return
            self._clear_detail_fields()
            self.marker_manager.select_marker(None)
            return
        marker = markers[0]
        self._pending_selection_id = None
        self._current_marker_id = marker.marker_id
        self._populate_detail_fields(marker, preserve_selection=True)
        self._selection_from_list = True
        try:
            self.marker_manager.select_marker(marker)
        finally:
            self._selection_from_list = False

    def _on_markers_changed(self, plane: str) -> None:
        if self._suspend_updates:
            return
        self._refresh_marker_list()
        current = self._current_marker()
        if current is not None:
            self._update_position_fields(current)
            self._ensure_marker_selected(current, preserve_existing=True)
        if self.placement_toggle.isChecked():
            self._configure_placement()

    def _selected_or_current_markers(self, marker_cls=None) -> List[Marker]:
        """Return selected markers, or fallback to current marker when appropriate."""
        if marker_cls is None:
            markers = list(self._selected_markers())
        else:
            markers = [marker for marker in self._selected_markers() if isinstance(marker, marker_cls)]
        if markers:
            return markers
        marker = self._current_marker()
        if marker is None:
            return []
        if marker_cls is not None and not isinstance(marker, marker_cls):
            return []
        return [marker]

    def _finalize_marker_style_update(
        self,
        changed_planes: Iterable[str],
        *,
        focus_marker_list: bool = False,
    ) -> None:
        changed = [str(plane).lower() for plane in changed_planes if plane]
        if not changed:
            return
        for plane in changed:
            self.marker_manager.redraw_plane(plane)
        self._notify_marker_state_changed(changed)
        self._refresh_marker_list()
        if focus_marker_list:
            self.marker_list.setFocus()

    def _update_placement_if_enabled(self) -> None:
        if self.placement_toggle.isChecked():
            self._configure_placement()

    def _notify_marker_state_changed(self, planes: Iterable[str]) -> None:
        """Notify observers that marker state changed without geometry edits."""
        seen: set[str] = set()
        for plane in planes:
            plane_name = str(plane or "").lower()
            if not plane_name or plane_name in seen:
                continue
            seen.add(plane_name)
            try:
                self.marker_manager.markers_changed.emit(plane_name)
            except Exception:
                continue

    def _on_manager_selection_changed(self, marker: Optional[Marker]) -> None:
        if self._suspend_selection_sync:
            return
        self._suspend_selection_sync = True
        try:
            if marker is None:
                if self._pending_selection_id:
                    self._schedule_pending_selection_restore()
                    return
                self.marker_list.clearSelection()
                self._clear_detail_fields()
                return
            self._pending_selection_id = None
            preserve_selection = self._selection_from_list
            self._ensure_marker_selected(
                marker,
                sync_manager=False,
                preserve_existing=preserve_selection,
            )
            self._populate_detail_fields(
                marker,
                preserve_selection=preserve_selection,
            )
        finally:
            self._suspend_selection_sync = False

    def _on_pixel_spin_changed(self, _value: float) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_markers()
        allow_multi = self._viewer_supports_position_edit(markers)
        if not allow_multi:
            marker = self._current_marker()
            if marker is None:
                return
            markers = [marker]
        elif not markers:
            return
        x_val = self.pixel_x_spin.value()
        y_val = self.pixel_y_spin.value()
        primary = markers[0]
        any_updated = False
        self._suspend_updates = True
        try:
            for target in markers:
                existing = list(target.pixel)
                if len(existing) < 2:
                    existing.extend([0.0] * (2 - len(existing)))
                if len(existing) >= 2 and math.isclose(existing[0], x_val, abs_tol=1e-6) and math.isclose(existing[1], y_val, abs_tol=1e-6):
                    continue
                existing[0] = x_val
                existing[1] = y_val
                self.marker_manager.update_marker_pixel(target, tuple(existing))
                any_updated = True
        finally:
            self._suspend_updates = False
        if any_updated:
            self._update_position_fields(primary)
            self._refresh_marker_list()
        else:
            self._update_position_fields(primary)

    def _on_pixel_entry_return_pressed(self) -> None:
        if self._suspend_updates:
            return
        if self._selected_markers() or self._current_marker() is not None:
            return
        if not getattr(self, "marker_manager", None):
            return
        plane = self._resolve_current_plane()
        marker = self._create_marker_from_coordinates(plane, self.pixel_x_spin.value(), self.pixel_y_spin.value())
        if marker is None:
            return
        self.marker_manager.redraw_plane(plane)

    def _on_world_entry_return_pressed(self) -> None:
        if self._suspend_updates:
            return
        if self._selected_markers() or self._current_marker() is not None:
            return
        if not getattr(self, "marker_manager", None):
            return
        world_inputs = (self.world_x_edit.text().strip(), self.world_y_edit.text().strip())
        if not all(world_inputs):
            return
        plane = self._resolve_current_plane()
        pixel_coords = self._world_inputs_to_pixels(plane, world_inputs)
        if pixel_coords is None:
            return
        marker = self._create_marker_from_coordinates(plane, pixel_coords[0], pixel_coords[1])
        if marker is None:
            return
        self.marker_manager.redraw_plane(plane)

    def _on_world_edit_finished(self) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_markers()
        allow_multi = self._viewer_supports_position_edit(markers)
        if not allow_multi:
            marker = self._current_marker()
            if marker is None:
                return
            markers = [marker]
        elif not markers:
            return
        primary = markers[0]
        world_inputs = (self.world_x_edit.text().strip(), self.world_y_edit.text().strip())
        if not all(world_inputs):
            self._update_position_fields(primary)
            return
        pixel_coords = self._world_inputs_to_pixels(primary.plane, world_inputs)
        if pixel_coords is None:
            self._update_position_fields(primary)
            return
        x_pix, y_pix = pixel_coords
        any_updated = False
        self._suspend_updates = True
        try:
            for target in markers:
                existing = list(target.pixel)
                if len(existing) < 2:
                    existing.extend([0.0] * (2 - len(existing)))
                if len(existing) >= 2 and math.isclose(existing[0], x_pix, abs_tol=1e-6) and math.isclose(existing[1], y_pix, abs_tol=1e-6):
                    continue
                existing[0] = x_pix
                existing[1] = y_pix
                self.marker_manager.update_marker_pixel(target, tuple(existing))
                any_updated = True
        finally:
            self._suspend_updates = False
        if any_updated:
            self._update_position_fields(primary)
            self._refresh_marker_list()
        else:
            self._update_position_fields(primary)

    def _on_world_frame_changed(self, _index: int) -> None:
        if self._suspend_updates:
            return
        current = self._current_marker()
        if current is not None:
            self._update_position_fields(current)
        self._refresh_marker_list()

    def _on_marker_geometry_changed(self, marker: Marker) -> None:
        if self._suspend_updates:
            return
        current = self._current_marker()
        if current is not None and marker.marker_id == current.marker_id:
            self._suspend_updates = True
            try:
                self._update_position_fields(marker)
                if isinstance(marker, LineMarker):
                    self._update_line_detail_fields(marker)
                elif isinstance(marker, TextMarker):
                    self._update_text_detail_fields(marker)
            finally:
                self._suspend_updates = False
        self._refresh_marker_list()

    def _on_type_changed(self, index: int) -> None:
        prev_kind = getattr(self, "_last_type_kind", None)
        programmatic = self._suspend_updates
        kind_data = self.TYPE_OPTIONS[index][1]
        kind = kind_data.get("kind")

        self._set_property_rows_visible(kind, kind_data.get("symbol"))

        self._update_label_field_visibility(kind)
        if kind == "text" and not self._selected_markers() and self._current_marker() is None:
            template_text = self._last_text_value or "Text"
            if self.label_edit.text() != template_text:
                self.label_edit.setText(template_text)

        if kind == "symbol":
            symbol = kind_data.get("symbol")
            if symbol in {"+", "x"} and self.symbol_controls.linewidth_spin.value() < 1.0:
                self.symbol_controls.linewidth_spin.setValue(1.0)

        if not programmatic:
            if prev_kind == "text" and kind != "text":
                self._last_text_value = self.label_edit.text().strip()
                self._suppress_label_from_text = True
                self.label_edit.clear()
            elif kind == "text":
                self._suppress_label_from_text = False

        if programmatic:
            self._last_type_kind = kind
            return

        changed_planes: set[str] = set()
        if kind == "symbol":
            markers = self._selected_or_current_markers(SymbolMarker)
            if markers:
                new_symbol = kind_data.get("symbol", "o")
                is_new_plus_cross = new_symbol in {"+", "x"}
                for target in markers:
                    style_dict = target.style.to_dict()
                    was_plus_cross = getattr(target, "symbol", None) in {"+", "x"}

                    if is_new_plus_cross and not was_plus_cross:
                        if style_dict.get("linewidth", 0.0) < 1.0:
                            style_dict["linewidth"] = 1.0
                        current_color = style_dict.get("color", "none")
                        if current_color != "none":
                            style_dict["edgecolor"] = current_color
                        elif style_dict.get("edgecolor", "none") == "none":
                            style_dict["edgecolor"] = "white"
                        style_dict["color"] = "none"
                    elif not is_new_plus_cross and was_plus_cross:
                        current_edge_color = style_dict.get("edgecolor", "none")
                        if current_edge_color != "none":
                            style_dict["color"] = current_edge_color
                            style_dict["edgecolor"] = "none"

                    target.symbol = new_symbol
                    target.update_style(MarkerStyle.from_dict(style_dict))
                    changed_planes.add(target.plane)

        if changed_planes:
            self._finalize_marker_style_update(changed_planes, focus_marker_list=True)

        self._capture_detail_group_height(kind)
        self._update_placement_if_enabled()

        self._last_type_kind = kind

    def _on_placement_toggled(self, enabled: bool) -> None:
        if enabled:
            self._configure_placement()
            if hasattr(self.viewer, "set_marker_mode"):
                self.viewer.set_marker_mode(True)
        else:
            self.marker_manager.cancel_placement()
            if hasattr(self.viewer, "set_marker_mode"):
                self.viewer.set_marker_mode(False)

    def _on_label_changed(self) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_or_current_markers()
        value = self.label_edit.text().strip()
        kind = self._current_kind()

        template_update_only = not markers

        if kind == "text":
            self._last_text_value = value or "Text"
        else:
            self._suppress_label_from_text = False

        if template_update_only:
            self._update_placement_if_enabled()
            return

        if markers:
            changed_planes: set[str] = set()
            for target in markers:
                if isinstance(target, TextMarker):
                    target.update_text(value)
                    changed_planes.add(target.plane)
                else:
                    target.set_label(value)
                    changed_planes.add(target.plane)
            self._finalize_marker_style_update(changed_planes)
            primary = markers[0]
            if isinstance(primary, TextMarker):
                self.label_edit.setText(primary.text)
        if kind == "text":
            if "$" in value:
                self.text_controls.font_combo.setCurrentText("STIXGeneral")
            else:
                if self.text_controls.font_combo.currentText() == "STIXGeneral":
                    fallback_font = next((font for font in self.FONT_OPTIONS if font != "STIXGeneral"), self.FONT_OPTIONS[0])
                    self.text_controls.font_combo.setCurrentText(fallback_font)

        self._update_placement_if_enabled()

    def _apply_common_style(self) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_or_current_markers()
        if not markers:
            self._update_placement_if_enabled()
            return

        color_value = self._normalized_color_value(self.color_combo.currentText())
        style_dict = {"opacity": self.opacity_spin.value()}

        changed_planes: set[str] = set()
        for target in markers:
            is_plus_cross = (
                isinstance(target, SymbolMarker) and getattr(target, "symbol", None) in {"+", "x"}
            )
            target_style = target.style.to_dict()
            target_style.update(style_dict)
            if is_plus_cross:
                target_style["edgecolor"] = color_value
                target_style["color"] = "none"
            else:
                target_style["color"] = color_value
            target.update_style(MarkerStyle.from_dict(target_style))
            changed_planes.add(target.plane)
        self._finalize_marker_style_update(changed_planes)
        self._update_placement_if_enabled()

    def _apply_symbol_style(self) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_or_current_markers(SymbolMarker)
        if not markers:
            self._update_placement_if_enabled()
            return
        changed_planes: set[str] = set()
        for target in markers:
            style_dict = target.style.to_dict()
            style_dict["size"] = self.symbol_controls.size_spin.value()
            linewidth_value = self.symbol_controls.linewidth_spin.value()
            symbol = getattr(target, "symbol", None)

            if symbol in {"+", "x"}:
                linewidth_value = max(1.0, linewidth_value)
                style_dict["edgecolor"] = self._normalized_color_value(
                    self.color_combo.currentText()
                )
                style_dict["color"] = "none"
            else:
                style_dict["edgecolor"] = self._normalized_color_value(
                    self.symbol_controls.edge_color_combo.currentText()
                )

            style_dict["linewidth"] = linewidth_value
            target.update_style(MarkerStyle.from_dict(style_dict))
            changed_planes.add(target.plane)
        self._finalize_marker_style_update(changed_planes)
        self._update_placement_if_enabled()

    def _apply_line_style(self) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_or_current_markers(LineMarker)
        if not markers:
            self._update_placement_if_enabled()
            return
        raw_length = self.line_controls.length_spin.value()
        unit = self.line_controls.unit_combo.currentText()
        linewidth_value = self.line_controls.linewidth_spin.value()
        color_value = self._normalized_color_value(self.color_combo.currentText())
        opacity_value = self.opacity_spin.value()
        style_mode = self.line_controls.style_combo.currentData()
        angle_value = self.line_controls.angle_spin.value()
        changed_planes: set[str] = set()
        for target in markers:
            pixels = self._convert_length_to_pixels(raw_length, unit, plane=target.plane)
            style_dict = target.style.to_dict()
            style_dict["linewidth"] = linewidth_value
            style_dict["color"] = color_value
            style_dict["opacity"] = opacity_value
            target.set_style_mode(style_mode)
            target.update_style(MarkerStyle.from_dict(style_dict))
            target.set_unit(unit)
            target.set_length(pixels, source_value=raw_length)
            target.set_angle(angle_value)
            changed_planes.add(target.plane)
        self._finalize_marker_style_update(changed_planes)
        self._update_placement_if_enabled()

    def _on_line_unit_changed(self, unit: str) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_or_current_markers(LineMarker)
        marker = markers[0] if markers else self._current_marker()
        if isinstance(marker, LineMarker):
            converted = self._convert_length_from_pixels(marker.length, unit, plane=marker.plane)
            self._suspend_updates = True
            try:
                self.line_controls.length_spin.setValue(converted)
            finally:
                self._suspend_updates = False
        self._apply_line_style()

    def _on_line_style_changed(self, _index: int) -> None:
        if self._suspend_updates:
            return
        self._apply_line_style()

    def _apply_text_style(self) -> None:
        if self._suspend_updates:
            return
        markers = self._selected_or_current_markers(TextMarker)
        if not markers:
            self._update_placement_if_enabled()
            return
        size_value = self.text_controls.font_size_spin.value()
        rotation_value = self.text_controls.rotation_spin.value()
        color_value = self._normalized_color_value(self.color_combo.currentText())
        font_value = self.text_controls.font_combo.currentText()
        opacity_value = self.opacity_spin.value()
        changed_planes: set[str] = set()
        for target in markers:
            style_dict = target.style.to_dict()
            style_dict["size"] = size_value
            style_dict["rotation"] = rotation_value
            text_value = target.text or ""
            style_dict["color"] = color_value
            if "$" in text_value:
                style_dict["font_family"] = "STIXGeneral"
            else:
                style_dict["font_family"] = font_value
            style_dict["opacity"] = opacity_value
            target.update_style(MarkerStyle.from_dict(style_dict))
            changed_planes.add(target.plane)
        self._finalize_marker_style_update(changed_planes)
        self._update_placement_if_enabled()

    def _on_delete_marker(self) -> None:
        markers = self._selected_markers()
        if not markers:
            marker = self._current_marker()
            if marker is not None:
                markers = [marker]
        if not markers:
            return
        planes: set[str] = set()
        for marker in markers:
            planes.add(marker.plane)
            self.marker_manager.remove_marker(marker.marker_id, marker.plane)
        self._refresh_marker_list()
        for plane in planes:
            self.marker_manager.redraw_plane(plane)

    def _on_delete_all(self) -> None:
        planes = list(self.marker_manager._layers.keys())
        if not planes:
            self._show_info("No markers to delete.")
            return
        any_removed = False
        for plane in planes:
            layer = self.marker_manager._layers.get(plane)
            if layer is None or not layer.markers:
                continue
            any_removed = True
            self.marker_manager.clear_plane(plane)
            self.marker_manager.redraw_plane(plane)
        if not any_removed:
            self._show_info("No markers to delete.")
            return
        self._current_marker_id = None
        self.marker_manager.select_marker(None)
        self._refresh_marker_list()

    def _on_deselect_markers(self) -> None:
        self._pending_selection_id = None
        self.marker_list.clearSelection()

    # ------------------------------------------------------------------
    # Persistence and export helpers
    def _choose_plane_for_export(self, title: str) -> Optional[str]:
        planes_set = {marker.plane for layer in self.marker_manager._layers.values() for marker in layer.markers.values()}
        def _plane_sort_key(name: str) -> Tuple[int, int, str]:
            base_order = {"xy": 0, "xz": 1, "zy": 2}.get(name, 50)
            suffix_num = 0
            try:
                parts = name.rsplit("_", 1)
                if len(parts) == 2:
                    suffix_num = int(parts[1])
            except Exception:
                suffix_num = 0
            return (base_order, suffix_num, name)
        planes = sorted(planes_set, key=_plane_sort_key)
        if not planes:
            return None
        multi_choice = len(planes) > 1
        options = list(planes)
        if multi_choice:
            options.insert(0, "all")
        if len(options) == 1:
            return options[0]
        plane_labels = [opt.upper() for opt in options]
        choice, ok = QInputDialog.getItem(self, title, "Select plane:", plane_labels, 0, False)
        if not ok or not choice:
            return None
        return options[plane_labels.index(choice)]

    def _on_save_json(self) -> None:
        plane = self._choose_plane_for_export("Save Markers")
        if plane is None:
            return
        if plane.lower() == "all":
            states: Dict[str, MarkerState] = {}
            for layer_plane in list(self.marker_manager._layers.keys()):
                states.update(self._collect_states(layer_plane))
            if not states:
                self._show_info("No markers to save.")
                return
        else:
            states = self._collect_states(plane)
            if not states:
                self._show_info(f"No markers to save for plane '{plane.upper()}'.")
                return
        default_dir = os.path.dirname(getattr(self.viewer, "filename_path", "")) or os.getcwd()
        file_plane = "all" if plane.lower() == "all" else plane
        filename, _ = QFileDialog.getSaveFileName(
            self,
            "Save Markers",
            os.path.join(default_dir, f"markers_{file_plane}.marker"),
            "Marker Files (*.marker *.marker.json *.json);;JSON Files (*.json);;All Files (*)",
        )
        if not filename:
            return
        if not filename.lower().endswith((".marker", ".marker.json", ".json")):
            filename += ".marker"
        try:
            world_frame = None
            if hasattr(self.marker_manager, "world_frame_for_plane"):
                try:
                    target_plane = next(iter(states.values())).plane if states else plane
                    world_frame = self.marker_manager.world_frame_for_plane(target_plane)
                except Exception:
                    world_frame = None
            payload = marker_states_to_json(states, plane=file_plane, world_frame=world_frame)
            with open(filename, "w", encoding="utf-8") as handle:
                handle.write(payload)
            self._show_info(f"Saved {len(states)} marker(s) to {os.path.basename(filename)}.")
        except Exception as exc:
            self._show_error(f"Failed to save markers: {exc}")

    def _on_load_json(self) -> None:
        default_dir = os.path.dirname(getattr(self.viewer, "filename_path", "")) or os.getcwd()
        path, _ = QFileDialog.getOpenFileName(
            self,
            "Load Markers",
            default_dir,
            "Marker Files (*.marker *.marker.json *.json);;All Files (*)",
        )
        if not path:
            return
        try:
            with open(path, "r", encoding="utf-8") as handle:
                payload = json.load(handle)
            plane = self.marker_manager.import_from_dict(payload)
            self.marker_manager.redraw_plane(plane)
            self._current_plane = plane
            self._refresh_marker_list()
            # self._show_info(f"Loaded markers for plane '{plane.upper()}'.")
        except Exception as exc:
            self._show_error(f"Failed to load markers: {exc}")

    def _on_export_ds9(self) -> None:
        plane = self._choose_plane_for_export("Export to DS9 Region")
        if plane is None:
            return
        states = self._collect_states(plane)
        if not states:
            self._show_info(f"No markers to export for plane '{plane.upper()}'.")
            return
        default_dir = os.path.dirname(getattr(self.viewer, "filename_path", "")) or os.getcwd()
        path, _ = QFileDialog.getSaveFileName(
            self,
            "Export Markers to DS9",
            os.path.join(default_dir, f"markers_{plane}.reg"),
            "DS9 Region Files (*.reg);;All Files (*)",
        )
        if not path:
            return
        if not path.lower().endswith(".reg"):
            path += ".reg"
        try:
            content = marker_states_to_ds9(states, plane)
            with open(path, "w", encoding="utf-8") as handle:
                handle.write(content)
            self._show_info(f"Exported DS9 region file to {os.path.basename(path)}.")
        except Exception as exc:
            self._show_error(f"Failed to export DS9 region file: {exc}")

    # ------------------------------------------------------------------
    # Placement configuration helpers
    def _configure_placement(self) -> None:
        if not self.placement_toggle.isChecked():
            return
        current_plane = self._resolve_current_plane()
        entry = self.TYPE_OPTIONS[self.type_combo.currentIndex()]
        config = dict(entry[1])
        kind = config.get("kind", "symbol")
        kwargs = dict(self._style_configuration_for_kind(kind))
        if kind == "symbol":
            kwargs.setdefault("symbol", config.get("symbol", "o"))
        self.marker_manager.begin_placement(kind, plane=current_plane, continuous=True, **kwargs)
        if hasattr(self.viewer, "set_marker_mode") and not getattr(self.viewer, "marker_mode_enabled", False):
            self.viewer.set_marker_mode(True)

    # ------------------------------------------------------------------
    # Utility helpers
    def _collect_states(self, plane: str) -> Dict[str, MarkerState]:
        try:
            self.marker_manager.refresh_world_coordinates(plane)
        except Exception:
            pass
        markers = self.marker_manager.markers_for_plane(plane)
        states = {marker.marker_id: marker.to_state() for marker in markers}

        # Attach world endpoints for lines so they can be reprojected with correct angle.
        viewer = self.marker_manager.viewer_for_plane(plane)
        format_pix = getattr(viewer, "format_pix", None) if viewer else None
        wcs = getattr(viewer, "wcs", None) if viewer else None
        base_plane_resolver = getattr(self.marker_manager, "_base_plane_for", None)
        base_plane = base_plane_resolver(plane) if callable(base_plane_resolver) else plane
        if format_pix is not None and wcs is not None:
            for marker in markers:
                if not isinstance(marker, LineMarker):
                    continue
                state = states.get(marker.marker_id)
                if state is None:
                    continue
                try:
                    start, end = marker._endpoints()
                    meta = dict(state.metadata)
                    wx0, wy0 = format_pix.pix_to_wcs(wcs, start[0], start[1], base_plane)
                    wx1, wy1 = format_pix.pix_to_wcs(wcs, end[0], end[1], base_plane)
                    meta["pixel_endpoints"] = [(float(start[0]), float(start[1])), (float(end[0]), float(end[1]))]
                    meta["world_endpoints"] = [(float(wx0), float(wy0)), (float(wx1), float(wy1))]
                    state.metadata = meta
                except Exception:
                    continue
        return states

    def _show_info(self, message: str) -> None:
        QMessageBox.information(self, "Markers", message)

    def _show_error(self, message: str) -> None:
        QMessageBox.warning(self, "Markers", message)

    def _pixel_scale_deg(self, plane: str) -> Optional[float]:
        resolver = getattr(self.marker_manager, "_pixel_scale_deg", None)
        if callable(resolver):
            try:
                scale = resolver(plane)
            except Exception:
                scale = None
            if scale:
                return float(scale)
        viewer = self._viewer_for_plane(plane)
        if viewer is None:
            return None
        wcs = getattr(viewer, "wcs", None)
        if wcs is None:
            return None
        axis_indices = self._plane_axis_indices(plane, wcs)
        if not axis_indices:
            axis_indices = (0, 1)
        try:
            scales = proj_plane_pixel_scales(wcs)
            if scales is not None and len(scales) >= 2:
                values = []
                for idx in axis_indices:
                    if idx < 0 or idx >= len(scales):
                        continue
                    val = scales[idx]
                    try:
                        values.append(abs(val.to_value(u.deg)))
                    except AttributeError:
                        try:
                            values.append(abs(float(val)))
                        except Exception:
                            continue
                if values:
                    avg = sum(values) / len(values)
                    if avg > 0:
                        return avg
        except Exception:
            pass
        try:
            cdelt = getattr(getattr(wcs, "wcs", None), "cdelt", None)
            if cdelt is not None and len(cdelt) > 0:
                values = []
                for idx in axis_indices:
                    if idx < 0 or idx >= len(cdelt):
                        continue
                    scale = abs(cdelt[idx])
                    if scale > 0:
                        values.append(float(scale))
                if values:
                    avg = sum(values) / len(values)
                    if avg > 0:
                        return avg
        except Exception:
            pass
        return None

    def _convert_length_to_pixels(self, value: float, unit: str, *, plane: str) -> float:
        if unit == "pixel":
            return float(value)
        scale_deg = self._pixel_scale_deg(plane)
        if not scale_deg:
            return float(value)
        if unit == "deg":
            return float(value) / scale_deg
        if unit == "arcmin":
            return float(value) / (scale_deg * 60.0)
        if unit == "arcsec":
            return float(value) / (scale_deg * 3600.0)
        return float(value)

    def _convert_length_from_pixels(self, value: float, unit: str, *, plane: str) -> float:
        if unit == "pixel":
            return float(value)
        scale_deg = self._pixel_scale_deg(plane)
        if not scale_deg:
            return float(value)
        if unit == "deg":
            return float(value) * scale_deg
        if unit == "arcmin":
            return float(value) * scale_deg * 60.0
        if unit == "arcsec":
            return float(value) * scale_deg * 3600.0
        return float(value)

    def _viewer_for_plane(self, plane: str):
        viewer_lookup = getattr(self.marker_manager, "viewer_for_plane", None)
        viewer = None
        if callable(viewer_lookup):
            try:
                viewer = viewer_lookup(plane)
            except Exception:
                viewer = None
        if viewer is None and getattr(self.viewer, "plane", None) == plane:
            viewer = self.viewer
        if viewer is None:
            has_plane = getattr(self.viewer, "has_marker_plane", None)
            if callable(has_plane):
                try:
                    if has_plane(plane):
                        viewer = self.viewer
                except Exception:
                    viewer = None
        return viewer or self.viewer

    # ------------------------------------------------------------------
    # Qt overrides
    def closeEvent(self, event) -> None:
        if self.placement_toggle.isChecked():
            self.placement_toggle.setChecked(False)
        super().closeEvent(event)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        self._on_deselect_markers()



import numpy as np
from typing import Optional

from PySide6.QtWidgets import (
    QWidget, QVBoxLayout, QLineEdit, QPushButton, QLabel,
    QHBoxLayout, QMessageBox, QComboBox, QFormLayout, QGroupBox,
)
from PySide6.QtCore import Qt
from takefits.ui.save_fits_dialog import SaveFITS
from datetime import datetime

# Import constants for Planck's law
from astropy import constants as const
from astropy import units as u
from astropy.io import fits
from takefits.core.usecases import convert_intensity_unit
from takefits.core.history_provenance import build_processing_history_lines
from takefits.tools.base_panel import (
    capture_preferred_cursor_snapshot,
    clear_action_preview_record,
    confirm_pending_close,
    has_action_record_tag,
    record_action_preview,
    replay_action_history_to_current_cursor,
)

class UnitConversionPanel(QWidget):
    """
    A panel for performing unit conversions
    (e.g., Jy/beam <-> K) for radio astronomy data cubes.
    
    It also allows editing key FITS header values (Rest Freq, BMAJ, BMIN, BPA).
    UI is based on QGroupBoxes, similar to masking.py.
    Size is fixed to prevent unwanted resizing.
    """
    
    def __init__(self, fits_viewer, subwindows):
        super().__init__()
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)
        self.fits_viewer = fits_viewer
        self.subwindows = subwindows
        self.original_data = None  # Defer copying data
        self.original_header = self.fits_viewer.header.copy() # Store original header for reset
        self.original_bunit = self._get_normalized_bunit(self.original_header)
        self._conversion_baseline_bunit = None
        self._conversion_baseline_has_bunit = False
        self.scaling_reference_data = None  # Baseline for manual scaling
        self._history_categories = {}  # Track latest history entries per category
        self._history_dedup_keys = set()
        self.spectral_metadata = getattr(self.fits_viewer, 'spectral_metadata', {})
        self._spectral_axis_index = self.spectral_metadata.get('axis_index')
        self._conversion_action_record_tag = "panel:unit-conversion"

        self._disk_header = self._load_disk_header()
        disk_axis_info = self._extract_disk_spectral_axis_info(self._disk_header)
        self._disk_spectral_axis_index = disk_axis_info.get('index')
        self._disk_spectral_ctype = disk_axis_info.get('ctype')
        self._disk_spectral_cunit = disk_axis_info.get('cunit')
        original_axis_type_meta = str(self.spectral_metadata.get('original_axis_type', '')).lower()
        if original_axis_type_meta in ('frequency', 'velocity'):
            self._disk_axis_was_frequency = (original_axis_type_meta == 'frequency')
        else:
            self._disk_axis_was_frequency = disk_axis_info.get('is_frequency', False)

        #current_axis_type_meta = str(self.spectral_metadata.get('current_axis_type', '')).lower()

        if getattr(self.fits_viewer.data, 'ndim', 0) <= 2:
            self.spectral_axis_mode = 'velocity'
        #elif current_axis_type_meta in ('frequency', 'velocity'):
        #    self.spectral_axis_mode = 'frequency' if current_axis_type_meta == 'frequency' else 'velocity'
        else:
            # 3D/4D cube: Default to the axis type of the *original* FITS file on disk
            self.spectral_axis_mode = 'frequency' if self._disk_axis_was_frequency else 'velocity'

        self._update_spectral_metadata_from_header()

        # --- Internal state for UI and calculations ---
        self.current_bunit = self.original_bunit
        self.header_modified = False # Track if header values have been changed
        
        # --- Check header for conversion capabilities ---
        self.header_params = self._check_header_keys()
        self.header_info_strings = self._get_header_info_strings() # Get strings for UI
        
        # --- Constants for calculations ---
        self.beam_area_sr = self._calculate_beam_area_sr()
        self.pixel_area_sr = self._calculate_pixel_area_sr()

        # Initialize UI elements
        self._init_ui_elements()
        self.initUI()
        
        # Set a fixed size based on the minimal initial layout
        self.setFixedSize(self.sizeHint())
        self.resync_after_workspace_restore()

    def _load_disk_header(self):
        filename_path = getattr(self.fits_viewer, 'filename_path', None)
        if not filename_path:
            return None
        try:
            return fits.getheader(filename_path)
        except Exception as err:
            print(f"Warning: Could not read original FITS header from disk: {err}")
            return None

    def _extract_disk_spectral_axis_info(self, header):
        info = {'index': None, 'ctype': None, 'cunit': None, 'is_frequency': False}
        if header is None:
            return info

        try:
            naxis = int(header.get('NAXIS', 0))
        except (TypeError, ValueError):
            naxis = 0

        for axis in range(1, naxis + 1):
            ctype = str(header.get(f'CTYPE{axis}', '')).strip()
            if not ctype:
                continue
            ctype_upper = ctype.upper()
            if 'FREQ' in ctype_upper:
                cunit = str(header.get(f'CUNIT{axis}', '')).strip()
                info.update(index=axis, ctype=ctype, cunit=cunit or 'Hz', is_frequency=True)
                return info

        for axis in range(1, naxis + 1):
            ctype = str(header.get(f'CTYPE{axis}', '')).strip()
            if not ctype:
                continue
            ctype_upper = ctype.upper()
            if any(tag in ctype_upper for tag in ('VRAD', 'VELO', 'VOPT')):
                cunit = str(header.get(f'CUNIT{axis}', '')).strip()
                info.update(index=axis, ctype=ctype, cunit=cunit or 'km/s', is_frequency=False)
                return info

        return info

    def _find_velocity_axis_in_header(self, header):
        """Return (axis_index, ctype, cunit) for a velocity-like axis if present."""
        info = (None, None, None)
        if header is None:
            return info

        try:
            naxis = int(header.get('NAXIS', 0))
        except (TypeError, ValueError):
            naxis = 0

        for axis in range(1, naxis + 1):
            ctype = str(header.get(f'CTYPE{axis}', '')).strip()
            if not ctype:
                continue
            ctype_upper = ctype.upper()
            if any(tag in ctype_upper for tag in ('VRAD', 'VELO', 'VOPT')):
                cunit = str(header.get(f'CUNIT{axis}', '')).strip()
                return axis, ctype, cunit

        return info

    @staticmethod
    def _classify_axis_mode(ctype):
        """Classify an axis CTYPE into 'frequency', 'velocity', or 'unknown'."""
        if not ctype:
            return 'unknown'
        ctype_upper = str(ctype).upper()
        if 'FREQ' in ctype_upper:
            return 'frequency'
        if any(tag in ctype_upper for tag in ('VRAD', 'VELO', 'VOPT')):
            return 'velocity'
        return 'unknown'

    def _update_spectral_metadata_from_header(self, header=None):
        """Synchronize shared spectral metadata with the latest header state."""
        if not isinstance(self.spectral_metadata, dict):
            return

        header = header or self.fits_viewer.header

        axis_index = (
            self._spectral_axis_index
            or self.spectral_metadata.get('axis_index')
            or self._disk_spectral_axis_index
        )

        if axis_index is not None:
            ctype = header.get(f'CTYPE{axis_index}', '')
            unit = header.get(f'CUNIT{axis_index}', '')
            mode = self._classify_axis_mode(ctype)
            self._spectral_axis_index = axis_index
            self.spectral_metadata['axis_index'] = axis_index
            self.spectral_metadata['current_axis_ctype'] = ctype
            self.spectral_metadata['current_axis_type'] = mode
            if isinstance(unit, str) and unit.strip():
                self.spectral_metadata['current_axis_unit'] = unit.strip()

        restfreq = header.get('RESTFRQ', header.get('RESTFREQ'))
        try:
            restfreq_value = float(restfreq)
        except (TypeError, ValueError):
            restfreq_value = None

        self.spectral_metadata['restfreq_hz'] = restfreq_value
        if self.spectral_metadata.get('restfreq_original_hz') is None:
            self.spectral_metadata['restfreq_original_hz'] = restfreq_value

    def _init_ui_elements(self):
        """Initialize UI widgets to allow access from multiple methods."""
        self.header_group = QGroupBox("Header Information (Freq/Beam/3rd Axis)")
        self.rf_input = QLineEdit()
        
        # Group for Beam Properties
        self.beam_group = QGroupBox("Beam Properties")
        self.bmaj_input = QLineEdit()
        self.bmin_input = QLineEdit()
        self.bpa_input = QLineEdit()

        self.set_button = QPushButton("Set")
        self.reset_header_button = QPushButton("Reset Header")
        self.spectral_axis_combo = QComboBox()
        self.spectral_axis_combo.addItem("Velocity", userData='velocity')
        self.spectral_axis_combo.addItem("Frequency", userData='frequency')

        # Group 3: Unit Conversion
        self.conversion_group = QGroupBox("Intensity Unit Conversion")
        self.current_unit_label = QLabel(self.current_bunit.upper())
        self.target_unit_combo = QComboBox(self)
        self.target_unit_combo.setFixedWidth(150)
        self.method_label = QLabel("Method:")
        self.method_combo = QComboBox(self)
        self.apply_conversion_button = QPushButton("Apply Unit Conversion")
        
        # Common buttons
        self.reset_scaling_button = QPushButton('Reset', self)
        self.reset_scaling_button.setEnabled(False)
        self.save_button = QPushButton('Save as FITS') 

        self.save_button.setEnabled(False) # Initially disabled

    def initUI(self):
        """
        Initialize the User Interface with QGroupBoxes for scaling modes.
        """
        main_layout = QVBoxLayout()

        # --- Build Group 3: Unit Conversion (Moved up)
        conversion_layout = QFormLayout(self.conversion_group)
        conversion_layout.setContentsMargins(5, 10, 5, 10)
        conversion_layout.setVerticalSpacing(5)
        conversion_layout.addRow(QLabel("Current Unit:"), self.current_unit_label)
        conversion_layout.addRow(QLabel("Target Unit:"), self.target_unit_combo)
        
        self.method_combo.addItems(["Rayleigh-Jeans", "Planck"])
        conversion_layout.addRow(self.method_label, self.method_combo)

        conversion_button_layout = QHBoxLayout()
        conversion_button_layout.addWidget(self.apply_conversion_button)
        conversion_button_layout.addStretch()
        conversion_button_layout.addWidget(self.reset_scaling_button)
        conversion_layout.addRow(conversion_button_layout)
        
        self.method_label.setEnabled(False)
        self.method_combo.setEnabled(False)
        
        main_layout.addWidget(self.conversion_group)

        # --- Build Group 2: Header Information ---
        header_layout = QFormLayout(self.header_group)
        header_layout.setContentsMargins(10, 10, 10, 10)
        header_layout.setVerticalSpacing(5) # Reduce vertical spacing
        header_layout.addRow(QLabel("RestFreq (GHz):"), self.rf_input)
        header_layout.addRow(QLabel("BMAJ (arcsec):"), self.bmaj_input)
        header_layout.addRow(QLabel("BMIN (arcsec):"), self.bmin_input)
        header_layout.addRow(QLabel("BPA (deg):"), self.bpa_input)
        header_layout.addRow(QLabel("Spectral Axis:"), self.spectral_axis_combo)

        header_button_layout = QHBoxLayout()
        header_button_layout.addWidget(self.set_button)
        header_button_layout.addStretch()
        header_button_layout.addWidget(self.reset_header_button)
        header_layout.addRow(header_button_layout)

        main_layout.addWidget(self.header_group)
        
        self._update_header_fields()
        self._populate_conversion_options()
        self._toggle_method_combo(self.target_unit_combo.currentText())
        self._sync_spectral_axis_ui()

        save_button_layout = QHBoxLayout()
        save_button_layout.addStretch(1)
        save_button_layout.addWidget(self.save_button)
        main_layout.addLayout(save_button_layout)

        main_layout.addStretch(1)


        self.setLayout(main_layout)
        self.setWindowTitle(f'Unit Convert: {self.fits_viewer.filename}')
        self.move_to_default_position()

        # Connect Signals
        self.apply_conversion_button.clicked.connect(self.apply_unit_conversion)
        self.reset_scaling_button.clicked.connect(self.reset_conversion_operations)
        self.save_button.clicked.connect(self.save_fits)
        self.target_unit_combo.currentTextChanged.connect(self._toggle_method_combo)
        
        self.set_button.clicked.connect(self.update_header_values)
        self.reset_header_button.clicked.connect(self.reset_header_values)
        self.rf_input.returnPressed.connect(self.update_header_values)
        self.bmaj_input.returnPressed.connect(self.update_header_values)
        self.bmin_input.returnPressed.connect(self.update_header_values)
        self.bpa_input.returnPressed.connect(self.update_header_values)
        self.spectral_axis_combo.currentIndexChanged.connect(self._on_spectral_axis_mode_changed)

    def move_to_default_position(self):
        """Move the panel next to the main control panel."""
        if hasattr(self.fits_viewer, 'control_panel'):
            control_panel_geometry = self.fits_viewer.control_panel.geometry()
            self.move(control_panel_geometry.x() + control_panel_geometry.width(), control_panel_geometry.y())

    # ------------------------------------------------------------------
    # Header and Capability Checks
    # ------------------------------------------------------------------

    def _get_normalized_bunit(self, header):
        """Get and normalize the BUNIT from the header."""
        bunit = header.get('BUNIT', 'unknown').strip().lower()
        if bunit == 'jy/beam':
            return 'jy/beam'
        elif bunit == 'k':
            return 'k'
        elif bunit in ('jy/pixel', 'jy/pix'):
            return 'jy/pix'
        return 'unknown'

    def _sync_bunit_to_viewers(self):
        """Keep live intensity unit labels synchronized with header BUNIT."""
        header = getattr(self.fits_viewer, 'header', None)
        bunit = ""
        if header is not None:
            try:
                bunit = str(header.get('BUNIT', '') or '').strip()
            except Exception:
                bunit = ""

        try:
            self.fits_viewer.bunit = bunit
        except Exception:
            pass

        for window in self.subwindows:
            if not window:
                continue
            try:
                window.bunit = bunit
            except Exception:
                continue

    def _check_header_keys(self):
        """
        Check the header for all keys required for conversions.
        Returns a dictionary of boolean flags.
        """
        header = self.fits_viewer.header # Always use the current header
        
        def _check_freq_axis(header):
            """Robust check for a valid frequency axis."""
            if 'RESTFRQ' in header:
                return True
            naxis = header.get('NAXIS', 0)
            for i in range(1, naxis + 1):
                ctype = header.get(f'CTYPE{i}', '')
                if 'FREQ' in ctype:
                    return True
                if 'VELO' in ctype or 'VRAD' in ctype:
                    break
            if naxis < 3 and 'CRVAL3' in header and 'FREQ' in header.get('CTYPE3', ''):
                 return True
            return False

        params = {
            'has_bmaj_bmin': 'BMAJ' in header and 'BMIN' in header,
            'has_cdelt': 'CDELT1' in header and 'CDELT2' in header,
            'has_freq_axis': _check_freq_axis(header)
        }
        
        self.can_convert_k = params['has_bmaj_bmin'] and params['has_freq_axis']
        self.can_convert_pix = params['has_bmaj_bmin'] and params['has_cdelt']
        
        return params
        
    def _get_header_info_strings(self):
        """Get header values as strings for the UI labels."""
        header = self.fits_viewer.header # Always use the current header
        info = {}
        
        try:
            info['bmaj'] = f"{header['BMAJ'] * 3600.0:.3f}"
            info['bmin'] = f"{header['BMIN'] * 3600.0:.3f}"
            info['bpa'] = f"{header['BPA']:.3f}"
        except Exception:
            info['bmaj'] = info.get('bmaj', "N/A")
            info['bmin'] = info.get('bmin', "N/A")
            info['bpa'] = info.get('bpa', "N/A")
            
        try:
            freq_hz = None
            if 'RESTFRQ' in header:
                freq_hz = header['RESTFRQ']
            else:
                naxis = header.get('NAXIS', 0)
                for i in range(1, naxis + 1):
                    ctype = header.get(f'CTYPE{i}', '')
                    if 'FREQ' in ctype:
                        freq_hz = header.get(f'CRVAL{i}')
                        break
                    if 'VELO' in ctype or 'VRAD' in ctype:
                        info['restfreq'] = "N/A"
                        break
            
            if freq_hz is not None:
                info['restfreq'] = f"{freq_hz / 1e9:.6f}"
            elif 'restfreq' not in info:
                info['restfreq'] = "N/A"
                
        except Exception:
            info['restfreq'] = "N/A (Error)"
            
        return info

    def _calculate_beam_area_sr(self):
        """Calculate beam solid angle in steradians. Returns 0 if not possible."""
        if not self.header_params['has_bmaj_bmin']:
            return 0.0
        try:
            bmaj_deg = self.fits_viewer.header['BMAJ']
            bmin_deg = self.fits_viewer.header['BMIN']
            bmaj_rad = np.deg2rad(bmaj_deg)
            bmin_rad = np.deg2rad(bmin_deg)
            return (np.pi * bmaj_rad * bmin_rad) / (4.0 * np.log(2.0))
        except Exception:
            return 0.0

    def _calculate_pixel_area_sr(self):
        """Calculate pixel solid angle in steradians. Returns 0 if not possible."""
        if not self.header_params['has_cdelt']:
            return 0.0
        try:
            cdelt1_deg = self.fits_viewer.header['CDELT1']
            cdelt2_deg = self.fits_viewer.header['CDELT2']
            cdelt1_rad = np.deg2rad(cdelt1_deg)
            cdelt2_rad = np.deg2rad(cdelt2_deg)
            return np.abs(cdelt1_rad * cdelt2_rad)
        except Exception:
            return 0.0

    # ------------------------------------------------------------------
    # UI Logic
    # ------------------------------------------------------------------

    def _update_header_fields(self):
        """Populate header input fields from current info strings."""
        self.header_info_strings = self._get_header_info_strings()
        self.rf_input.setText(self.header_info_strings.get('restfreq', 'N/A'))
        self.bmaj_input.setText(self.header_info_strings.get('bmaj', 'N/A'))
        self.bmin_input.setText(self.header_info_strings.get('bmin', 'N/A'))
        self.bpa_input.setText(self.header_info_strings.get('bpa', 'N/A'))
        self._sync_spectral_axis_ui()

    def _set_field_set_style(self, field: QLineEdit, is_set: bool):
        """Update the style and read-only state of a field to show if it's set."""
        field.setReadOnly(is_set)
        if is_set:
            field.setStyleSheet("background-color: #e6ffed; color: #555;") # Light green, grey text
        else:
            field.setStyleSheet("") # Default style

    def _populate_conversion_options(self):
        """
        Populate the target unit combo box based on the current unit,
        disable options if header keys are missing, and set a default target.
        """
        self.target_unit_combo.clear() 
        
        default_target_userdata = None
        
        if self.current_bunit == 'jy/beam':
            self.target_unit_combo.addItem("K (Brightness Temp)", userData='k')
            self.target_unit_combo.addItem("Jy/pixel", userData='jy/pix')
            self._set_option_enabled('k', self.can_convert_k)
            self._set_option_enabled('jy/pix', self.can_convert_pix)
            if self.can_convert_k: default_target_userdata = 'k'
            elif self.can_convert_pix: default_target_userdata = 'jy/pix'
                
        elif self.current_bunit == 'k':
            self.target_unit_combo.addItem("Jy/beam", userData='jy/beam')
            self.target_unit_combo.addItem("Jy/pixel", userData='jy/pix')
            self._set_option_enabled('jy/beam', self.can_convert_k)
            self._set_option_enabled('jy/pix', self.can_convert_k and self.can_convert_pix)
            if self.can_convert_k: default_target_userdata = 'jy/beam'
            elif self.can_convert_k and self.can_convert_pix: default_target_userdata = 'jy/pix'

        elif self.current_bunit == 'jy/pix':
            self.target_unit_combo.addItem("Jy/beam", userData='jy/beam')
            self.target_unit_combo.addItem("K (Brightness Temp)", userData='k')
            self._set_option_enabled('jy/beam', self.can_convert_pix)
            self._set_option_enabled('k', self.can_convert_pix and self.can_convert_k)
            if self.can_convert_pix and self.can_convert_k: default_target_userdata = 'k'
            elif self.can_convert_pix: default_target_userdata = 'jy/beam'

        if default_target_userdata:
            index_to_select = self.target_unit_combo.findData(default_target_userdata)
            if index_to_select != -1:
                self.target_unit_combo.setCurrentIndex(index_to_select)

        has_enabled_option = any(self.target_unit_combo.model().item(i).isEnabled() for i in range(self.target_unit_combo.count()))

        # Overall check for enabling the conversion group
        can_convert = has_enabled_option and self.current_bunit != 'unknown'
        self.target_unit_combo.setEnabled(can_convert)
        self.apply_conversion_button.setEnabled(can_convert)
        # The method combo's enabled state is handled by _toggle_method_combo

        if not can_convert:
            tooltip = "Current unit (BUNIT) is unknown or unsupported." if self.current_bunit == 'unknown' else "Required FITS header keys are missing or not set."
            self.conversion_group.setToolTip(tooltip)
            self.target_unit_combo.setCurrentIndex(-1)
            self.method_label.setEnabled(False)
            self.method_combo.setEnabled(False)
        else:
            self.conversion_group.setToolTip("")
            # After enabling, specifically check if method combo should be shown
            self._toggle_method_combo(self.target_unit_combo.currentText())

    def _sync_spectral_axis_ui(self):
        """Update spectral axis combo state based on data dimensionality and mode."""
        if not hasattr(self, 'spectral_axis_combo'):
            return

        if getattr(self.fits_viewer.data, 'ndim', 0) <= 2:
            self.spectral_axis_combo.setEnabled(False)
            self.spectral_axis_combo.setToolTip("Spectral axis options require 3D/4D data.")
            self.spectral_axis_combo.blockSignals(True)
            self.spectral_axis_combo.setCurrentIndex(0)
            self.spectral_axis_combo.blockSignals(False)
            return

        self.spectral_axis_combo.setEnabled(True)
        self.spectral_axis_combo.setToolTip("")

        desired_mode = getattr(self, 'spectral_axis_mode', 'velocity')
        index = self.spectral_axis_combo.findData(desired_mode)
        if index != -1 and index != self.spectral_axis_combo.currentIndex():
            self.spectral_axis_combo.blockSignals(True)
            self.spectral_axis_combo.setCurrentIndex(index)
            self.spectral_axis_combo.blockSignals(False)

    def _on_spectral_axis_mode_changed(self, _index: int):
        """Handle spectral axis selection changes."""
        if not self.spectral_axis_combo.isEnabled():
            return

        mode = self.spectral_axis_combo.currentData()
        if mode not in ('velocity', 'frequency'):
            return

        if mode == getattr(self, 'spectral_axis_mode', None):
            return

        self.spectral_axis_mode = mode
        mode_label = "Velocity (VRAD)" if mode == 'velocity' else "Frequency (FREQ)"
        self._update_history(f"Spectral axis preference set to {mode_label}.", category='spectral-axis')
        self.header_modified = True
        self.save_button.setEnabled(True)

    def _set_option_enabled(self, userData, enabled):
        """Finds item by userData and enables/disables it."""
        for i in range(self.target_unit_combo.count()):
            if self.target_unit_combo.itemData(i) == userData:
                item = self.target_unit_combo.model().item(i)
                if item:
                    item.setEnabled(enabled)
                return

    def _toggle_method_combo(self, text):
        """Enable/Disable the method combo box based on selection."""
        # First, check if the parent group is enabled at all
        if not self.target_unit_combo.isEnabled():
            self.method_label.setEnabled(False)
            self.method_combo.setEnabled(False)
            return

        target_unit = self.target_unit_combo.currentData()
        is_k_conversion = (target_unit == 'k' or self.current_bunit == 'k')
        show_method = is_k_conversion and self.can_convert_k

        self.method_label.setEnabled(show_method)
        self.method_combo.setEnabled(show_method)
        
        if show_method:
            is_3d_plus = self.fits_viewer.header.get('NAXIS', 2) >= 3
            planck_item = self.method_combo.model().item(1)
            if planck_item:
                planck_item.setEnabled(is_3d_plus)
            self.method_combo.setToolTip("" if is_3d_plus else "Planck conversion requires a 3D or 4D cube.")
            if not is_3d_plus:
                self.method_combo.setCurrentIndex(0)
        else:
            self.method_combo.setToolTip("")

    # ------------------------------------------------------------------
    # Main Scaling and Header Update Logic
    # ------------------------------------------------------------------

    def _ensure_original_data(self):
        """Check if original_data is cached, if not, cache it."""
        if self.original_data is None:
            self.original_data = self.fits_viewer.data.copy()
            header = getattr(self.fits_viewer, 'header', None)
            self._conversion_baseline_has_bunit = bool(header is not None and 'BUNIT' in header)
            self._conversion_baseline_bunit = header.get('BUNIT') if header is not None else None


    def update_header_values(self):
        """
        Validate and apply new header values from the UI.
        For each changed field, lock it and change its style upon success.
        """
        updated_keys = []
        error_messages = []
        ignore_values = ['n/a', 'nan', 'none', 'non', '']
        detail_history_messages = []

        def process_field(field, key, conversion_factor, is_degree=False):
            # Skip if the field is already read-only (already set)
            if field.isReadOnly():
                return

            text = field.text().strip()
            original_text = self.header_info_strings.get(key.lower() if not is_degree else 'bpa', 'N/A')
            
            if text.lower() in ignore_values:
                if original_text.lower() not in ignore_values:
                     # User deleted a valid value, treat as reset for this key
                    if key in self.original_header:
                        self.fits_viewer.header[key] = self.original_header[key]
                    elif key in self.fits_viewer.header:
                        del self.fits_viewer.header[key]
                    updated_keys.append(key)
                return # No change or reset to original, don't lock

            if text == original_text:
                return # No change

            try:
                value = float(text)
                # Special handling for RESTFRQ to log original and adjust velocity axis headers
                if key == 'RESTFRQ':
                    old_has = 'RESTFRQ' in self.fits_viewer.header
                    old_val_hz = self.fits_viewer.header.get('RESTFRQ', None)
                    new_val_hz = value * conversion_factor # This is a float

                    # --- BUG FIX (from previous turn) ---
                    # Attempt to convert old_val_hz to a float for comparison
                    old_val_hz_float = None
                    if old_val_hz is not None:
                        try:
                            old_val_hz_float = float(old_val_hz)
                        except (TypeError, ValueError):
                            pass # old_val_hz_float remains None

                    # Check if the value is *actually* changing
                    is_changed = (old_val_hz_float is None) or (not np.isclose(old_val_hz_float, new_val_hz))
                    # --- BUG FIX (from previous turn) END ---

                    self.fits_viewer.header['RESTFRQ'] = new_val_hz
                    if isinstance(self.spectral_metadata, dict):
                        try:
                            old_val_hz_numeric = float(old_val_hz)
                        except (TypeError, ValueError):
                            old_val_hz_numeric = None
                        self.spectral_metadata['restfreq_hz'] = new_val_hz
                        if self.spectral_metadata.get('restfreq_original_hz') is None:
                            original_rest = old_val_hz_numeric if old_has and old_val_hz_numeric is not None else new_val_hz
                            self.spectral_metadata['restfreq_original_hz'] = original_rest
                    
                    # HISTORY: record precise change
                    if old_has:
                        try:
                            old_val_hz_fmt = f"{float(old_val_hz):.6f}"
                        except (TypeError, ValueError):
                            old_val_hz_fmt = str(old_val_hz)
                        
                        # *** APPLY BUG FIX CHECK HERE ***
                        if is_changed:
                            detail_history_messages.append((
                                'restfreq-change',
                                f"RESTFRQ changed: {old_val_hz_fmt} Hz -> {new_val_hz:.6f} Hz"
                            ))
                    else:
                        # If it didn't exist before, it's definitely a change
                        detail_history_messages.append((
                            'restfreq-change',
                            f"RESTFRQ set to {new_val_hz:.6f} Hz (previously undefined)"
                        ))
                        is_changed = True # Ensure 'is_changed' is true if key was new

                    # If spectral axis is velocity-like, keep type and ensure units
                    naxis = self.fits_viewer.header.get('NAXIS', 0)
                    for i in range(1, naxis + 1):
                        ctype_key = f'CTYPE{i}'
                        ctype_i = self.fits_viewer.header.get(ctype_key, '')
                        ctype_upper = ctype_i.upper()
                    
                        current_spectral_axis = self._spectral_axis_index
                        if current_spectral_axis is None:
                            break
                        if i != current_spectral_axis:
                            continue
                        
                        if any(tag in ctype_upper for tag in ('VELO', 'VRAD', 'VOPT')):
                            normalized_ctype = ctype_i if ctype_i else 'VRAD'
                            self.fits_viewer.header[ctype_key] = normalized_ctype
                            cunit_key = f'CUNIT{i}'
                            cunit_val = str(self.fits_viewer.header.get(cunit_key, '')).strip()
                            if not cunit_val:
                                self.fits_viewer.header[cunit_key] = 'km/s'
                                cunit_val = 'km/s'

                            # Only run CRVAL update logic if the frequency *actually* changed
                            if old_has and old_val_hz and is_changed and old_val_hz_float is not None:
                                try:
                                    velocity_value = float(self.fits_viewer.header.get(f'CRVAL{i}', 0.0))
                                except (TypeError, ValueError):
                                    velocity_value = None

                                if velocity_value is not None:
                                    c_speed = const.c.to('m/s').value
                                    vel_unit = cunit_val.lower()
                                    vel_mps = velocity_value * 1000.0 if 'km/s' in vel_unit else velocity_value
                                    if np.isfinite(vel_mps):
                                        freq_obs = None
                                        if 'VOPT' in ctype_upper:
                                            denom = 1.0 + vel_mps / c_speed
                                            if np.isfinite(denom) and denom > 0:
                                                freq_obs = old_val_hz_float / denom
                                        else:
                                            freq_obs = old_val_hz_float * (1.0 - vel_mps / c_speed)

                                        if freq_obs and np.isfinite(freq_obs) and freq_obs > 0:
                                            if 'VOPT' in ctype_upper:
                                                new_vel_mps = c_speed * ((new_val_hz / freq_obs) - 1.0)
                                            else:
                                                # Use new_val_hz (which is guaranteed > 0 if freq_obs > 0)
                                                new_vel_mps = c_speed * (new_val_hz - freq_obs) / new_val_hz

                                            if np.isfinite(new_vel_mps):
                                                new_vel = new_vel_mps / 1000.0 if 'km/s' in vel_unit else new_vel_mps
                                                self.fits_viewer.header[f'CRVAL{i}'] = new_vel
                                                crval_key = f'CRVAL{i}'
                                                if crval_key not in updated_keys:
                                                    updated_keys.append(crval_key)
                            break
                    
                    # Only add RESTFRQ to updated_keys and lock the field IF the value actually changed.
                    if is_changed:
                        if key not in updated_keys:
                            updated_keys.append(key)
                        self._set_field_set_style(field, True) # Lock and style the field

                else:
                    self.fits_viewer.header[key] = value * conversion_factor
                    
                    # Moved these lines inside the 'else' block
                    if key not in updated_keys:
                        updated_keys.append(key)
                    self._set_field_set_style(field, True) # Lock and style the field

            except ValueError:
                error_messages.append(f"{key} is not a valid number.")
            except Exception as e:
                error_messages.append(f"Error setting {key}: {e}")

        process_field(self.rf_input, 'RESTFRQ', 1e9)
        process_field(self.bmaj_input, 'BMAJ', 1/3600.0)
        process_field(self.bmin_input, 'BMIN', 1/3600.0)
        process_field(self.bpa_input, 'BPA', 1.0, is_degree=True)

        # --- Finalize ---
        if error_messages:
            QMessageBox.warning(self, 'Input Error', '\n'.join(error_messages))
            # Revert fields that failed validation
            self.reset_header_values(show_message=False)
            return

        if not updated_keys:
            QMessageBox.information(self, 'No Changes', 'No valid header values were changed.')
        else:
            self.header_modified = True
            self.save_button.setEnabled(True)
            
            for key in updated_keys:
                for window in self.subwindows:
                    if window:
                        window.header[key] = self.fits_viewer.header[key]
            
            history_msg = f"Header updated: {', '.join(updated_keys)} modified by user."
            self._update_history(history_msg, category='header-update')
            for category, message in detail_history_messages:
                self._update_history(message, category=category, include_general=False)

            # Recalculate parameters and update UI state
            self._reinitialize_parameters()
            # Immediately propagate changes to viewers
            self.update_all_displays()
            self._update_spectral_metadata_from_header()
            
            QMessageBox.information(self, 'Success', f"FITS header has been updated for: {', '.join(updated_keys)}.")


    def _reinitialize_parameters(self):
        """Recalculate all internal parameters after a header change."""
        self.header_params = self._check_header_keys()
        # Do not update text fields here, only the underlying capabilities
        self.beam_area_sr = self._calculate_beam_area_sr()
        self.pixel_area_sr = self._calculate_pixel_area_sr()
        self._populate_conversion_options()

    def apply_unit_conversion(self):
        """Apply the selected unit conversion."""
        self._ensure_original_data()

        try:
            target_unit = self.target_unit_combo.currentData()
            if target_unit is None:
                QMessageBox.warning(self, 'Input Error', 'Please select a valid target unit.')
                return

            method_str = 'Rayleigh-Jeans' if 'Rayleigh-Jeans' in self.method_combo.currentText() else 'Planck'
            
            # Map method name to usecase expectation ("rayleigh-jeans", "planck")
            uc_method = method_str.lower()
            from_unit = self._get_normalized_bunit(self.fits_viewer.header)
            self.current_bunit = from_unit
            
            scaled_data, new_bunit_header = convert_intensity_unit(
                data=self.fits_viewer.data,
                header=self.fits_viewer.header,
                from_unit=from_unit,
                to_unit=target_unit,
                method=uc_method
            )
            
            # Record History (Simplified compared to old implementation)
            history_msg = f"Unit conversion: {self.current_bunit.upper()} -> {new_bunit_header} (Method:{method_str})"
            
            # We can't easily get the min/max factor without re-doing the math, so we skip detail logging of the factor.
            # This is acceptable for Tier 2 migration.
            
            self._apply_data_and_history(scaled_data, new_bunit_header, history_msg, history_category='conversion')
            self._record_conversion_preview_action(from_unit, target_unit, uc_method)

            self.current_bunit = new_bunit_header.lower()
            self.current_unit_label.setText(new_bunit_header)
            self.conversion_group.setToolTip("Reset panel to perform another conversion.")
            for widget in [self.target_unit_combo, self.apply_conversion_button, self.method_label, self.method_combo]:
                widget.setEnabled(False)
            self.reset_scaling_button.setEnabled(True)
            self.save_button.setEnabled(True)
            # Reset manual scaling baseline after conversions
            self.scaling_reference_data = None

        except Exception as e:
            QMessageBox.critical(self, 'Conversion Error', f'An error occurred: {e}')
            self.reset_conversion_operations()



    def _apply_data_and_history(self, scaled_data, new_bunit_header, history_msg, history_category=None):
        """Helper function to apply data, header, history, and update displays."""
        self._update_history(history_msg, category=history_category)

        self.fits_viewer.header['BUNIT'] = new_bunit_header
        self.fits_viewer.data = scaled_data
        self.fits_viewer.update_cube()
        
        for window in self.subwindows:
            if window:
                window.header['BUNIT'] = new_bunit_header
                window.data = scaled_data
                window.update_cube()
        self._sync_bunit_to_viewers()
        self.update_all_displays()

    def reset_header_values(self, show_message=True):
        """Resets header values to their original state, making fields editable again."""
        if self.header_modified and show_message:
            QMessageBox.information(self, 'Header Reset', 'Header values have been reset to their original state.')

        self.fits_viewer.header = self.original_header.copy()
        for window in self.subwindows:
            if window:
                window.header = self.original_header.copy()
        
        # Unlock all fields and reset their styles
        for field in [self.rf_input, self.bmaj_input, self.bmin_input, self.bpa_input]:
            self._set_field_set_style(field, False)

        # Repopulate fields with original values and re-check capabilities
        self._update_header_fields()
        self._reinitialize_parameters()
        if getattr(self.fits_viewer.data, 'ndim', 0) > 2:
            self.spectral_axis_mode = 'frequency' if self._disk_axis_was_frequency else 'velocity'
        self._sync_spectral_axis_ui()
        for category in ('spectral-axis', 'restfreq-change', 'header-update'):
            self._clear_history_category(category)
        self._history_dedup_keys.clear()
        
        self.header_modified = False
        
        # Disable save button ONLY if no data scaling has been applied
        if self.original_data is None:
            self.save_button.setEnabled(False)

        # Ensure viewers redraw with restored metadata
        self._sync_bunit_to_viewers()
        self.update_all_displays()
        self._update_spectral_metadata_from_header()

    def reset_conversion_operations(self):
        """Resets only the unit conversion UI to its original state."""
        preferred_cursor = capture_preferred_cursor_snapshot(self.fits_viewer)
        removed_preview = self._clear_conversion_preview_action()
        restored_from_history = False
        can_restore_from_history = removed_preview and (not self.header_modified)
        if can_restore_from_history:
            restored_from_history = replay_action_history_to_current_cursor(
                self.fits_viewer,
                preferred_cursor=preferred_cursor,
            )

        # Restore data if it was changed
        if not restored_from_history and self.original_data is not None:
            self.fits_viewer.data = self.original_data
            self.fits_viewer.update_cube()
            if self._conversion_baseline_has_bunit:
                self.fits_viewer.header['BUNIT'] = self._conversion_baseline_bunit
            else:
                self.fits_viewer.header.pop('BUNIT', None)
            for window in self.subwindows:
                if window:
                    window.data = self.original_data
                    if self._conversion_baseline_has_bunit:
                        window.header['BUNIT'] = self._conversion_baseline_bunit
                    else:
                        window.header.pop('BUNIT', None)
                    window.update_cube()
        self.original_data = None  # Nullify cache
        self._conversion_baseline_bunit = None
        self._conversion_baseline_has_bunit = False

        # Restore internal state and UI for conversion
        self._sync_bunit_to_viewers()
        self.current_bunit = self._get_normalized_bunit(self.fits_viewer.header)
        self.original_bunit = self.current_bunit
        self.current_unit_label.setText(self.current_bunit.upper().replace('JY','Jy').replace('PIX','pixel'))
        
        # Only disable save if header hasn't also been modified
        if not self.header_modified:
            self.save_button.setEnabled(False)

        # Re-initialize conversion UI components
        self.conversion_group.setToolTip("")
        for widget in [self.target_unit_combo, self.apply_conversion_button, self.method_label, self.method_combo]:
            widget.setEnabled(True)
        self.reset_scaling_button.setEnabled(False)
        self._populate_conversion_options() # This re-runs the enable/disable logic

        # Update displays
        self.update_all_displays()
        self._clear_history_category('conversion')
        self._update_spectral_metadata_from_header()

    def _record_conversion_preview_action(self, from_unit: str, to_unit: str, method: str) -> None:
        record_action_preview(
            self.fits_viewer,
            "convert_intensity_unit",
            {
                "from_unit": str(from_unit),
                "to_unit": str(to_unit),
                "method": str(method),
            },
            replace_tag=self._conversion_action_record_tag,
        )

    def _clear_conversion_preview_action(self) -> bool:
        return bool(
            clear_action_preview_record(
                self.fits_viewer,
                self._conversion_action_record_tag,
                action_name="convert_intensity_unit",
            )
        )

    def update_all_displays(self):
        """Force-redraw all viewer and subwindow canvases with updated data."""
        self._sync_bunit_to_viewers()
        all_windows = [self.fits_viewer] + self.subwindows
        for window in all_windows:
            if not window:
                continue

            window.update_cube()

            data_slice = None
            data = getattr(window, 'data', None)
            current_channel = 0
            if hasattr(window, 'current_channel_index'):
                try:
                    current_channel = window.current_channel_index()
                except Exception:
                    current_channel = 0
            plane = getattr(window, 'plane', 'xy')
            if data is None:
                continue

            try:
                if data.ndim == 4:
                    if plane == 'xy':
                        data_slice = data[0, current_channel, :, :]
                    elif plane == 'xz':
                        data_slice = data[0, :, current_channel, :]
                    elif plane == 'zy':
                        data_slice = data[0, :, :, current_channel].T
                elif data.ndim == 3:
                    if plane == 'xy':
                        data_slice = data[current_channel, :, :]
                    elif plane == 'xz':
                        data_slice = data[:, current_channel, :]
                    elif plane == 'zy':
                        data_slice = data[:, :, current_channel].T
                elif data.ndim == 2:
                    data_slice = data
            except Exception:
                data_slice = None

            if data_slice is not None and hasattr(window, 'im'):
                window.im.set_data(data_slice)
            if hasattr(window, 'canvas'):
                window.canvas.draw()

        main_window = getattr(self.fits_viewer, 'main_window', None) or self.fits_viewer
        refresh_wcs_strings = getattr(main_window, '_refresh_wcs_display_strings', None)
        if callable(refresh_wcs_strings):
            try:
                refresh_wcs_strings()
            except Exception:
                pass
                
    # ------------------------------------------------------------------
    # FITS Save and HISTORY
    # ------------------------------------------------------------------
  
    def _prepare_header_for_save(self, *, is_data_being_flipped_back: bool = False):
        """Return a header copy adjusted for FITS output."""
        header_copy = self.fits_viewer.header.copy()

        convert_to_frequency = (
            getattr(self.fits_viewer.data, 'ndim', 0) > 2
            and getattr(self, 'spectral_axis_mode', 'velocity') == 'frequency'
        )
        
        axis = self._spectral_axis_index or self._disk_spectral_axis_index
        if axis is None:
             axis, _, _ = self._find_velocity_axis_in_header(header_copy)

        if convert_to_frequency:
            
            desired_unit = None
            desired_ctype = None

            if axis is not None:
                restfreq = header_copy.get('RESTFRQ', header_copy.get('RESTFREQ'))
                try:
                    restfreq = float(restfreq)
                except (TypeError, ValueError):
                    restfreq = None

                crval_key = f'CRVAL{axis}'
                cdelt_key = f'CDELT{axis}'
                cunit_key = f'CUNIT{axis}'
                ctype_key = f'CTYPE{axis}'

                # Use the *original* axis type (from disk) to determine the scenario
                original_was_frequency = self._disk_axis_was_frequency
                
                # (Scenario 1: Freq -> Freq)
                if original_was_frequency:
                    
                    # (Freq -> Freq) means we restore the *original* spectral axis
                    # from the disk header, ignoring the current (Velo) axis.
                    
                    original_disk_header = self._load_disk_header()
                    if original_disk_header:
                         # Restore the *spectral axis* keys from disk.
                         for key_prefix in ('CTYPE', 'CUNIT', 'CRVAL', 'CRPIX', 'CDELT'):
                             key = f"{key_prefix}{axis}"
                             if key in original_disk_header:
                                 header_copy[key] = original_disk_header[key]
                             elif key in header_copy:
                                 # If key existed in viewer but not disk, remove it
                                 if key_prefix not in ('CRPIX'): # Keep CRPIX if needed?
                                     del header_copy[key]
                    
                    # Ensure CTYPE/CUNIT are valid if disk_header failed
                    current_ctype = header_copy.get(ctype_key, '')
                    if not current_ctype or 'VELO' in current_ctype.upper():
                        header_copy[ctype_key] = self._disk_spectral_ctype or 'FREQ'
                        
                    current_cunit = header_copy.get(cunit_key, '')
                    if not current_cunit or 'm/s' in current_cunit.lower():
                        header_copy[cunit_key] = self._disk_spectral_cunit or 'Hz'
                    
                    # Ensure the *new* RESTFRQ (from the UI) is in the header
                    if restfreq is not None:
                        header_copy['RESTFRQ'] = restfreq
                    
                    return header_copy
                
                # (Scenario 3: Velo -> Freq)
                # If we reach here, it means:
                # convert_to_frequency=True
                # AND original_was_frequency=False
                
                # Load the *target* format.
                if self.spectral_metadata.get('original_axis_type') == 'frequency':
                    desired_unit = self.spectral_metadata.get('original_axis_unit')
                    desired_ctype = self.spectral_metadata.get('original_axis_ctype')

                if desired_unit is None and self._disk_spectral_cunit:
                    desired_unit = self._disk_spectral_cunit
                if desired_ctype is None and self._disk_spectral_ctype and 'FREQ' in self._disk_spectral_ctype.upper():
                    desired_ctype = self._disk_spectral_ctype

                # Ensure desired_unit is a valid frequency unit
                desired_unit = (desired_unit or '').strip()
                try:
                    unit_obj = u.Unit(desired_unit) if desired_unit else None
                    if unit_obj is None or not unit_obj.is_equivalent(u.Hz):
                        desired_unit = 'Hz'
                except Exception:
                    desired_unit = 'Hz'

                desired_ctype = desired_ctype or 'FREQ'


                if restfreq and np.isfinite(restfreq):
                    vel_crval = header_copy.get(crval_key, 0.0)
                    vel_cdelt = header_copy.get(cdelt_key, 0.0)
                    vel_cunit = str(header_copy.get(cunit_key, '') or '').strip().lower()
                    if not vel_cunit:
                        vel_cunit = str(self.spectral_metadata.get('current_axis_unit', '') or '').strip().lower()

                    try:
                        vel_crval = float(vel_crval)
                    except (TypeError, ValueError):
                        vel_crval = None

                    try:
                        vel_cdelt = float(vel_cdelt)
                    except (TypeError, ValueError):
                        vel_cdelt = None

                    c_mps = const.c.to('m/s').value
                    vel_unit_factor = 1000.0 if vel_cunit == 'km/s' else 1.0

                    freq_crval = None
                    freq_cdelt = None
                    
                    current_ctype_upper = str(header_copy.get(ctype_key, '')).upper()

                    if vel_crval is not None:
                        vel_crval_mps = vel_crval * vel_unit_factor
                        if 'VOPT' in current_ctype_upper:
                            denom = 1.0 + vel_crval_mps / c_mps
                            if np.isfinite(denom) and denom != 0.0:
                                freq_crval = restfreq / denom
                        else:
                            freq_crval = restfreq * (1.0 - vel_crval_mps / c_mps)

                    if vel_cdelt is not None:
                        vel_cdelt_mps = vel_cdelt * vel_unit_factor
                        freq_cdelt = - (vel_cdelt_mps * restfreq) / c_mps

                    desired_unit = desired_unit or 'Hz'
                    try:
                        target_unit = u.Unit(desired_unit)
                        if not target_unit.is_equivalent(u.Hz):
                            target_unit = u.Hz
                            desired_unit = 'Hz'
                    except Exception:
                        target_unit = u.Hz
                        desired_unit = 'Hz'

                    if freq_crval is not None:
                        freq_crval = (freq_crval * u.Hz).to(target_unit).value
                        header_copy[crval_key] = freq_crval
                    if freq_cdelt is not None:
                        freq_cdelt = (freq_cdelt * u.Hz).to(target_unit).value
                        header_copy[cdelt_key] = freq_cdelt

                    if desired_unit:
                        header_copy[cunit_key] = desired_unit

                    header_copy[ctype_key] = desired_ctype or 'FREQ'
        
        elif axis is not None: # Target is Velocity (convert_to_frequency is False)
            
            # Scenarios (Velo -> Velo) and (Freq -> Velo) are handled here.
            # In both cases, the loader and UI have already ensured the in-memory
            # header (which we have in header_copy) correctly describes the data
            # with a velocity axis, including any rest frequency changes.
            # We just need to ensure the CTYPE and CUNIT are present and valid.

            ctype_key = f'CTYPE{axis}'
            cunit_key = f'CUNIT{axis}'

            desired_ctype = 'VRAD' 
            desired_unit = 'km/s'
            
            # Try to get more specific types from metadata if available
            if self.spectral_metadata.get('current_axis_type') == 'velocity':
                desired_ctype = self.spectral_metadata.get('current_axis_ctype') or desired_ctype
                desired_unit = self.spectral_metadata.get('current_axis_unit') or desired_unit
            elif self.spectral_metadata.get('original_axis_type') == 'velocity':
                # Fallback to original if current is not available
                desired_ctype = self.spectral_metadata.get('original_axis_ctype') or desired_ctype
                desired_unit = self.spectral_metadata.get('original_axis_unit') or desired_unit

            header_copy[ctype_key] = header_copy.get(ctype_key) or desired_ctype
            header_copy[cunit_key] = header_copy.get(cunit_key) or desired_unit
            
            # The complex re-calculation logic that was here for the Freq->Velo
            # case was buggy. It incorrectly mixed original header values with the new
            # rest frequency, ignoring the already-correct in-memory header state.
            # By simply trusting the in-memory header, we fix the bug.

        return header_copy

    def save_fits(self):
        """Save the currently displayed (scaled) data to a new FITS file."""
        if self.original_data is None and not self.header_modified:
            QMessageBox.information(self, 'No Changes', 'No scaling or header changes have been applied to save.')
            return

        # Determine suffix based on changes
        suffix_parts = []
        
        # 1. Check for intensity unit conversion
        bunit_changed = self.original_data is not None
        if bunit_changed:
            if self.current_bunit == 'k':
                suffix_parts.append("Tb")
            elif self.current_bunit == 'jy/beam':
                suffix_parts.append("Jybeam")
            elif self.current_bunit == 'jy/pix':
                suffix_parts.append("Jypix")

        # 2. Check for spectral axis conversion
        axis_mode_changed = (
            getattr(self.fits_viewer.data, 'ndim', 0) > 2 and
            self.header_modified and 
            self.spectral_axis_combo.isEnabled()
        )
        if axis_mode_changed:
            current_mode = self.spectral_axis_mode
            original_mode_is_freq = self._disk_axis_was_frequency
            
            if current_mode == 'frequency' and not original_mode_is_freq:
                suffix_parts.append("freq")
            elif current_mode == 'velocity' and original_mode_is_freq:
                suffix_parts.append("vel")

        # 3. Construct final suffix
        if suffix_parts:
            suffix = "_".join(suffix_parts)
        else: # This branch is taken if header was modified but no specific suffix applies
            suffix = "hdr_edit"

        # Prepare a copy of the data to be potentially flipped back
        data_to_save = self.fits_viewer.data.copy()

        # Check if the data should be reverted to its original order
        was_flipped_on_load = self.spectral_metadata.get('axis_flipped', False)
        was_originally_freq = self._disk_axis_was_frequency
        is_frequency_mode = (getattr(self, 'spectral_axis_mode', 'velocity') == 'frequency')
        revert_flip = was_flipped_on_load and was_originally_freq and is_frequency_mode

        if revert_flip:
            # If the data was flipped on load (for internal consistency) and originated
            # from a frequency axis, revert it to its original order for saving.
            if self._spectral_axis_index is not None and data_to_save.ndim >= 3:
                spectral_axis_dim = data_to_save.ndim - self._spectral_axis_index
                
                if 0 <= spectral_axis_dim < data_to_save.ndim:
                    data_to_save = np.flip(data_to_save, axis=spectral_axis_dim)
                    print("\033[96mData array reverted to original FITS file order for saving.\033[0m")

        valid_data = data_to_save[np.isfinite(data_to_save)]
        data_min = float(np.nanmin(valid_data)) if valid_data.size > 0 else 0.0
        data_max = float(np.nanmax(valid_data)) if valid_data.size > 0 else 0.0

        new_header = self._prepare_header_for_save(is_data_being_flipped_back=revert_flip)
        new_header['DATAMIN'] = data_min
        new_header['DATAMAX'] = data_max

        stale_history_lines = []
        for lines in self._history_categories.values():
            stale_history_lines.extend(str(line) for line in lines if line)
        self._remove_history_lines(new_header, stale_history_lines)

        for entry in build_processing_history_lines(self.fits_viewer):
            new_header.add_history(entry)

        original_header_for_save = self._disk_header or self.fits_viewer.header
        save_fits = SaveFITS(data_to_save, new_header, self.fits_viewer.filename, original_header=original_header_for_save)
        save_fits.save(suffix=suffix)

    def _remove_history_lines(self, header, lines):
        """Remove specific HISTORY lines from a header."""
        if not header or not lines:
            return
        try:
            existing_entries = list(header.get('HISTORY', []))
        except KeyError:
            existing_entries = []

        if not existing_entries:
            return

        targets = {line for line in lines if line}
        if not targets:
            return

        filtered_entries = [entry for entry in existing_entries if entry not in targets]

        # Remove all HISTORY cards and re-add filtered ones to preserve order
        if 'HISTORY' in header:
            header.remove('HISTORY', remove_all=True, ignore_missing=True)

        for entry in filtered_entries:
            header.add_history(entry)

    def _clear_history_category(self, category: str):
        """Remove the stored HISTORY entries for a given category."""
        previous_lines = self._history_categories.pop(category, None)
        if not previous_lines:
            return
        self._remove_history_lines(self.fits_viewer.header, previous_lines)
        for window in self.subwindows:
            if window and window.header:
                self._remove_history_lines(window.header, previous_lines)
        self._history_dedup_keys = {key for key in self._history_dedup_keys if key[0] != category}

    @staticmethod
    def _split_lines(text, limit=70):
        """Split text into chunks that fit into FITS HISTORY cards."""
        if not text:
            return []
        return [text[i:i+limit] for i in range(0, len(text), limit)]

    def _update_history(self, operation_details: str, category: Optional[str] = None, include_general: bool = True):
        """Adds a formatted HISTORY entry to the FITS header.

        When a category is provided, only the latest entry per category is kept.
        An entry with the same category and details will not be added if it already exists.
        """
        if not (self.fits_viewer and self.fits_viewer.header):
            return

        dedup_key = (category, operation_details)
        if dedup_key in self._history_dedup_keys:
            return

        if category:
            self._clear_history_category(category)

        timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        
        # Prepare full strings
        raw_operation_entry = f"Operation details: {operation_details}"
        raw_general_entry = (
            f"Data scaled/header updated using takefits on {timestamp}"
            if include_general else None
        )

        # Split into safe chunks (list of strings)
        operation_chunks = self._split_lines(raw_operation_entry)
        general_chunks = self._split_lines(raw_general_entry) if raw_general_entry else []

        headers_to_update = []
        main_header = getattr(self.fits_viewer, 'header', None)
        if main_header:
            headers_to_update.append(main_header)
        for window in self.subwindows:
            header = getattr(window, 'header', None) if window else None
            if header:
                headers_to_update.append(header)

        seen_headers = set()
        for header in headers_to_update:
            header_id = id(header)
            if header_id in seen_headers:
                continue
            seen_headers.add(header_id)
            self._insert_history_entries(header, general_chunks, operation_chunks, category)

        self._history_dedup_keys.add(dedup_key)

        if category:
            entries_to_store = tuple(general_chunks + operation_chunks)
            if entries_to_store:
                self._history_categories[category] = entries_to_store

    def _insert_history_entries(self, header, general_chunks: list, operation_chunks: list, category):
        """Insert history entries with controlled ordering and de-duplication."""
        if not header:
            return

        try:
            entries = list(header.get('HISTORY', []))
        except KeyError:
            entries = []

        modified = False

        # Add General Entry
        if general_chunks:
            if general_chunks[0] not in entries:
                entries.extend(general_chunks)
                modified = True

        # Add Operation Entry
        if operation_chunks:
            if operation_chunks[0] not in entries:
                insert_index = self._determine_history_insert_index(entries, category)
                
                if insert_index >= len(entries):
                    entries.extend(operation_chunks)
                else:
                    for chunk in reversed(operation_chunks):
                        entries.insert(insert_index, chunk)
                modified = True

        if modified:
            if 'HISTORY' in header:
                header.remove('HISTORY', remove_all=True, ignore_missing=True)
            for entry in entries:
                header.add_history(entry)

    def _determine_history_insert_index(self, entries, category: Optional[str]) -> int:
        """Determine where to insert a new operation entry."""
        # Find index of the most recent general entry
        general_idx = self._find_history_index(entries, "Data scaled/header updated using takefits on ")

        if category == 'header-update':
            return general_idx + 1 if general_idx is not None else len(entries)

        if category == 'restfreq-change':
            # Look for the START of a "Header updated" entry. 
            header_idx = self._find_history_index(entries, "Operation details: Header updated:")
            if header_idx is not None:
                return header_idx + 1
            return general_idx + 1 if general_idx is not None else len(entries)

        if general_idx is not None:
            return general_idx + 1
        return len(entries)

    @staticmethod
    def _find_history_index(entries, prefix: str) -> Optional[int]:
        """Return the index of the last entry starting with the given prefix."""
        for idx in range(len(entries) - 1, -1, -1):
            if entries[idx].startswith(prefix):
                return idx
        return None

    # ------------------------------------------------------------------
    # Close Event
    # ------------------------------------------------------------------

    def _has_pending_close_changes(self) -> bool:
        """Return True if this panel has unapplied conversion/header changes."""
        current_bunit = self._get_normalized_bunit(self.fits_viewer.header)
        has_preview = bool(has_action_record_tag(self.fits_viewer, self._conversion_action_record_tag))
        return (
            self.original_data is not None
            or self.header_modified
            or current_bunit != self.original_bunit
            or has_preview
        )

    def _discard_pending_close_changes(self) -> None:
        """Discard panel changes and restore original data/header state."""
        if self.original_data is not None or has_action_record_tag(self.fits_viewer, self._conversion_action_record_tag):
            self.reset_conversion_operations()
        current_bunit = self._get_normalized_bunit(self.fits_viewer.header)
        if self.header_modified or current_bunit != self.original_bunit:
            self.reset_header_values(show_message=False)

    def resync_after_workspace_restore(self) -> None:
        if self._has_pending_close_changes():
            self.save_button.setEnabled(True)

    def closeEvent(self, event):
        """Handle window close event."""
        if self._has_pending_close_changes():
            choice = confirm_pending_close(
                self,
                "Close Unit Conversion Panel",
                "There are unapplied unit conversion/header changes.",
            )
            if choice == "cancel":
                event.ignore()
                return
            if choice == "discard":
                self._discard_pending_close_changes()
        super().closeEvent(event)

    # ------------------------------------------------------------------
    # Conversion Calculation Helpers
    # ------------------------------------------------------------------

    # ------------------------------------------------------------------
    # Conversion Calculation Helpers (Deprecated - Moved to Core/UseCases)
    # ------------------------------------------------------------------
    # Methods _get_freq_axis_hz, _reshape_freqs_for_broadcast, and unit conversions
    # have been migrated to core.usecases.convert_intensity_unit.
